import json
import math
from neuron import h


class Chk:
    """A test helper that compares stored results with test results
    chk = Chk(filename)

    chk("name", value) normally asserts that the value == dict[name]
    But, if dict[name] does not exist, the value is added to the dict.
    """

    def __init__(self, fname, must_exist=True):
        """JSON with name, fname, must be a dict.
        Assert file exists by default.
        A temporary must_exist=False provides a simple way for the test
        author to bootstrap the file.
        """
        import os

        self.fname = fname
        self.d = {}
        self.modified = False
        if must_exist:
            assert os.path.exists(fname)
        else:
            if not os.path.exists(fname):
                return

        with open(fname, "rb") as f:
            self.d = json.load(f)
        assert type(self.d) is dict

    def __call__(self, key, value, tol=0.0):
        """Assert value == dict[key] unless value is a hoc_Vector.
        For the Vector case, assert the maximum relative element difference is <= tol.
        If the key does not exist add {key:value} to the dict.
        """

        if key in self.d:
            if type(value) == type(h.Vector):  # actually hoc.HocObject
                # Convert to list to keep the `equal` method below simple
                value = list(value)
            # Hand-rolled comparison that uses `tol` for arithmetic values
            # buried inside lists of lists.
            def equal(a, b):
                assert type(a) == type(b)
                if type(a) in (float, int):
                    match = math.isclose(a, b, rel_tol=tol)
                    if not match:
                        print(a, b, "diff", abs(a - b) / max(abs(a), abs(b)), ">", tol)
                    return match
                elif type(a) == str:
                    return a == b
                elif type(a) == list:
                    # List comprehension avoids short-circuit, so the "diff"
                    # message just above is printed for all elements
                    return all([equal(aa, bb) for aa, bb in zip(a, b)])
                raise Exception(
                    "Don't know how to compare objects of type " + str(type(a))
                )

            match = equal(value, self.d[key])
            if not match:
                print(key, "difference")
                print("std = ", self.d[key])
                print("val = ", value)
            assert match
        else:
            print("{} added {}".format(self, key))
            if type(value) == type(h.Vector):  # actually hoc.HocObject
                self.d[key] = value.to_python()
            else:
                self.d[key] = value
            self.modified = True

    def rm(self, key):
        """Remove key from dict.
        Only needed temporarily when a key value needs updating.
        """
        if key in self.d:
            del self.d[key]
            self.modified = True

    def save(self):
        """If dict has been modified, save back to original filename
        This is typically called at the end of the test and normally
        does nothing unless the author adds a new chk(...) to the test.
        """
        if self.modified:
            with open(self.fname, "w") as f:
                json.dump(self.d, f, indent=2)
                f.write("\n")



from neuron import h, rxd
from math import pi

h.load_file("stdrun.hoc")
h.CVode().active(True)
sec = h.Section()
# the extracellular space
ecs = rxd.Extracellular(
    -55, -55, -55, 55, 55, 55, dx=10, volume_fraction=0.2, tortuosity=1.6
)

# Who?
x = rxd.Species([ecs], name="x", d=0.0, charge=1, initial=0)


def callbackfun():
    return 1000


x[ecs].node_by_location(-20, 0, 0).include_flux(1000)

x[ecs].node_by_location(0, 0, 0).include_flux(callbackfun)

x[ecs].node_by_location(20, 0, 0).include_flux(sec(0.5)._ref_v)

h.finitialize(1000)
h.CVode().event(10)
h.continuerun(10)

print("float", x[ecs].node_by_location(-20, 0, 0).concentration)
print("callb", x[ecs].node_by_location(0, 0, 0).concentration)
print("refer", x[ecs].node_by_location(20, 0, 0).concentration)



from neuron import h, gui
from neuron.units import ms, mV

h.load_file("stdrun.hoc")


class Cell:
    def __init__(self, gid, x, y, z, theta):
        self._gid = gid
        self._setup_morphology()
        self.all = self.soma.wholetree()
        self._setup_biophysics()
        self.x = self.y = self.z = 0
        h.define_shape()
        self._rotate_z(theta)
        self._set_position(x, y, z)

        # everything below here in this method is NEW
        self._spike_detector = h.NetCon(self.soma(0.5)._ref_v, None, sec=self.soma)
        self.spike_times = h.Vector()
        self._spike_detector.record(self.spike_times)

        self._ncs = []

        self.soma_v = h.Vector().record(self.soma(0.5)._ref_v)

    def __repr__(self):
        return "{}[{}]".format(self.name, self._gid)

    def _set_position(self, x, y, z):
        for sec in self.all:
            for i in range(sec.n3d()):
                sec.pt3dchange(
                    i,
                    x - self.x + sec.x3d(i),
                    y - self.y + sec.y3d(i),
                    z - self.z + sec.z3d(i),
                    sec.diam3d(i),
                )
        self.x, self.y, self.z = x, y, z

    def _rotate_z(self, theta):
        """Rotate the cell about the Z axis."""
        for sec in self.all:
            for i in range(sec.n3d()):
                x = sec.x3d(i)
                y = sec.y3d(i)
                c = h.cos(theta)
                s = h.sin(theta)
                xprime = x * c - y * s
                yprime = x * s + y * c
                sec.pt3dchange(i, xprime, yprime, sec.z3d(i), sec.diam3d(i))


class BallAndStick(Cell):
    name = "BallAndStick"

    def _setup_morphology(self):
        self.soma = h.Section(name="soma", cell=self)
        self.dend = h.Section(name="dend", cell=self)
        self.dend.connect(self.soma)
        self.soma.L = self.soma.diam = 12.6157
        self.dend.L = 200
        self.dend.diam = 1

    def _setup_biophysics(self):
        for sec in self.all:
            sec.Ra = 100  # Axial resistance in Ohm * cm
            sec.cm = 1  # Membrane capacitance in micro Farads / cm^2
        self.soma.insert("hh")
        for seg in self.soma:
            seg.hh.gnabar = 0.12  # Sodium conductance in S/cm2
            seg.hh.gkbar = 0.036  # Potassium conductance in S/cm2
            seg.hh.gl = 0.0003  # Leak conductance in S/cm2
            seg.hh.el = -54.3  # Reversal potential in mV
        # Insert passive current in the dendrite
        self.dend.insert("pas")
        for seg in self.dend:
            seg.pas.g = 0.001  # Passive conductance in S/cm2
            seg.pas.e = -65  # Leak reversal potential mV

        # NEW: the synapse
        self.syn = h.ExpSyn(self.dend(0.5))
        self.syn.tau = 2 * ms



from neuron import h, crxd as rxd, gui

# reaction rates
kf = 0.02
kb = 0.01
# initial concentrations
ca0 = 1
b0 = 0.5
cab0 = 0.001
# diffusion coefficients
Dca = 1
Db = 0
Dcab = 0

# create sections
soma = h.Section(name="soma")
dend = h.Section(name="dend")
soma.L = 100
soma.diam = 1
soma.nseg = 101
dend.L = 100
dend.diam = 1
dend.nseg = 101
soma.connect(dend)


r = rxd.Region(h.allsec())
Ca = rxd.Species(r, name="Ca", d=Dca)
Buf = rxd.Species(r, name="Buf", d=Db)
CaBuf = rxd.Species(r, name="CaBuf", d=Dcab)

buffering = rxd.Reaction(Ca + Buf, CaBuf, kf, kb)


#
# set initial concentrations to ca0, b0, cab0 in soma,
# and to 0.001 in dend
#
Ca.initial = lambda node: (ca0 if node.sec == soma else 0.001)
Buf.initial = lambda node: (b0 if node.sec == soma else 0.001)
CaBuf.initial = lambda node: (cab0 if node.sec == soma else 0.001)


h.init()
species_0_trace = h.Vector()
species_0_trace.record(Ca.nodes(soma)(0.5)[0]._ref_concentration)

species_1_trace = h.Vector()
species_1_trace.record(Buf.nodes(soma)(0.5)[0]._ref_concentration)

species_2_trace = h.Vector()
species_2_trace.record(CaBuf.nodes(soma)(0.5)[0]._ref_concentration)


times = h.Vector()
times.record(h._ref_t)
#
# simulate 1 sec
#
h.CVode().active(1)
h.CVode().re_init()
h.CVode().atol(1e-12)
h.tstop = 1000
h.run()

if __name__ == "__main__":
    from matplotlib import pyplot

    #
    # plot
    #
    labels = ["Ca", "Buf", "CaBuf"]
    for i, trace in enumerate([species_0_trace, species_1_trace, species_2_trace]):
        pyplot.plot(times.as_numpy(), trace.as_numpy(), label=labels[i])
    pyplot.legend()
    pyplot.title("Ca+Buf->CaBuf")
    pyplot.xlabel("Time (ms)")
    pyplot.ylabel("Concentration (mM)")
    pyplot.tight_layout()
    pyplot.show()



import weakref
from . import rxdmath, rxd, node, species, region, initializer
import numpy
from .generalizedReaction import (
    GeneralizedReaction,
    get_scheme_rate1_rate2_regions_custom_dynamics_mass_action,
)
from .constants import molecules_per_mM_um3
from neuron import h
import itertools
from .rxdException import RxDException


def _ref_list_with_mult(obj):
    result = []
    for i, p in zip(list(obj.keys()), list(obj.values())):
        w = weakref.ref(i)
        result += [w] * p
    return result


class MultiCompartmentReaction(GeneralizedReaction):
    def __init__(self, *args, **kwargs):
        """Specify a reaction spanning multiple regions to be added to the system.

        Use this for, for example, pumps and channels, or interactions between
        species living in a volume (e.g. the cytosol) and species on a
        membrane (e.g. the plasma membrane).

        For each species/state/parameter, you must specify what region you are
        referring to, as it could be present in multiple regions. You must
        also specify a `membrane` or a `border` (these are treated as synonyms)
        that separates the regions involved in your reaction. This is necessary
        because the default behavior is to scale the reaction rate by the
        border area, as would be expected if one of the species involved is a
        pump that is binding to a species in the volume. If this is not the
        desired behavior, pass the keyword argument `scale_by_area=False`.

        Pass in `membrane_flux=True` if the reaction produces a current across
        the plasma membrane that should affect the membrane potential.

        Unlike Reaction objects, the base units for the rates are in terms of
        molecules per square micron per ms.

        .. seealso::

            :class:`neuron.rxd.Reaction`
        """

        # parse the arguments shared with rxd.Reaction
        (
            scheme,
            rate_f,
            rate_b,
            regions,
            custom_dynamics,
            mass_action,
        ) = get_scheme_rate1_rate2_regions_custom_dynamics_mass_action(args, kwargs)
        # additional keyword arguments
        membrane_flux = kwargs.get("membrane_flux", False)
        membrane = kwargs.get("membrane")
        border = kwargs.get("border")
        scale_by_area = kwargs.get("scale_by_area", True)

        if border is not None and membrane is not None:
            raise Exception("cannot specify both border and membrane; they are synoyms")
        if border is not None:
            membrane = border

        # TODO: verify schemes use weakrefs
        self._scheme = scheme
        self._scale_by_area = scale_by_area
        self._original_rate_f = rate_f
        self._original_rate_b = rate_b

        for ar in [scheme, rate_f, rate_b]:
            try:
                if ar._voltage_dependent:
                    self._voltage_dependent = True
                    break
            except AttributeError:
                pass
        else:
            self._voltage_dependent = False

        if custom_dynamics is not None and mass_action is not None:
            raise RxDException("Cannot specify both custom_dynamics and mass_action.")
        elif custom_dynamics is None and mass_action is None:
            custom_dynamics = False
        elif custom_dynamics is None and mass_action is not None:
            custom_dynamics = not mass_action
        self._custom_dynamics = custom_dynamics
        self._trans_membrane = True
        if membrane_flux not in (True, False):
            raise RxDException("membrane_flux must be either True or False")
        if membrane is None:
            raise RxDException("MultiCompartmentReaction requires a membrane parameter")
        if membrane_flux and species._has_3d:
            raise RxDException("membrane_flux not supported except in 1D")
        self._membrane_flux = membrane_flux
        if not isinstance(scheme, rxdmath._Reaction):
            raise RxDException("%r not a recognized reaction scheme" % self._scheme)
        self._dir = scheme._dir
        if not membrane._geometry.is_area():
            raise RxDException("must specify a membrane not a volume for the boundary")
        self._regions = [membrane]
        rxd._register_reaction(self)

        # initialize self if the rest of rxd is already initialized
        if initializer.is_initialized():
            self._do_init()
            self._update_indices()

    def _do_init(self):
        self._update_rates()

    def _update_rates(self):
        lhs = self._scheme._lhs._items
        rhs = self._scheme._rhs._items
        rate_f = self._original_rate_f
        rate_b = self._original_rate_b
        if not self._custom_dynamics:
            for k, v in zip(list(lhs.keys()), list(lhs.values())):
                if v == 1:
                    rate_f = rate_f * k
                else:
                    rate_f = rate_f * k**v
        if rate_b is not None:
            if self._dir in ("<", ">"):
                raise RxDException(
                    "unidirectional Reaction can have only one rate constant"
                )
            if not self._custom_dynamics:
                for k, v in zip(list(rhs.keys()), list(rhs.values())):
                    if v == 1:
                        rate_b = rate_b * k
                    else:
                        rate_b = rate_b * k**v
            rate = rate_f - rate_b
            self._sources = _ref_list_with_mult(lhs)
            self._dests = _ref_list_with_mult(rhs)
        elif self._dir == "<>":
            raise RxDException("bidirectional Reaction needs two rate constants")
        elif self._dir == ">":
            rate = rate_f
            self._sources = _ref_list_with_mult(lhs)
            self._dests = _ref_list_with_mult(rhs)
        elif self._dir == "<":
            self._sources = _ref_list_with_mult(rhs)
            self._dests = _ref_list_with_mult(lhs)
            rate = rate_f
        else:
            raise RxDException("unrecognized direction; should never happen")

        # check that the regions have sections
        for reg in self._regions:
            if any(reg._secs1d) or any(reg._secs3d):
                break
        else:
            return

        # check for 3D sections
        self._src3d = set()
        self._dst3d = set()
        self._mem3d = set(self._regions[0]._secs3d)
        sources = [
            s()._region()
            for s in self._sources
            if not isinstance(s(), species.SpeciesOnExtracellular)
        ]
        dests = [
            s()._region()
            for s in self._dests
            if not isinstance(s(), species.SpeciesOnExtracellular)
        ]
        for reg in sources:
            if any(reg._secs3d):
                self._src3d.update(reg._secs3d)
        for reg in dests:
            if any(reg._secs3d):
                self._dst3d.update(reg._secs3d)
        # if self._src3d.intersection(self._dst3d).intersection(self._mem3d):
        #    #Find all interacting voxels for each grid. Also build up the 2D array of index maps
        #    raise RxDException('Multicompartment reactions in 3D are not yet supported.')

        # check there are common 1D section between source and destination
        mem1d = set(self._regions[0]._secs1d)
        src1d = set()
        dst1d = set()
        for reg in sources:
            if any(reg._secs1d):
                src1d.update(reg._secs1d)
        for reg in dests:
            if any(reg._secs1d):
                dst1d.update(reg._secs1d)
        if sources:
            mem1d = mem1d.intersection(src1d)
        if dests:
            mem1d = mem1d.intersection(dst1d)
        if (src1d or dst1d) and not mem1d:
            msg = (
                "sources and destinations"
                if sources and dests
                else "sources"
                if sources
                else "destinations"
            )
            raise RxDException(
                "Multicompartment reactions the membrane and %s must share common sections. %r"
                % (msg, self)
            )

        self._changing_species = list(set(self._sources + self._dests))

        mem = self._regions[0]
        regs = [mem]
        # for sptr in self._changing_species:
        #    if isinstance(sptr(), species.Species):
        #        raise RxDException('must specify region for all involved species')
        #    elif hasattr(sptr(),'_extracellular'):
        #        regs.append(sptr()._extracellular()._region)
        #    else:
        #        regs.append(sptr()._region())
        self._rate, self._involved_species = rxdmath._compile(rate, regs)

    @property
    def f_rate(self):
        return self._original_rate_f

    @property
    def b_rate(self):
        return self._original_rate_b

    @f_rate.setter
    def f_rate(self, value):
        if self._dir not in ("<>", ">"):
            raise RxDException("no forward reaction in reaction scheme")
        self._original_rate_f = value
        self._update_rates()

    @b_rate.setter
    def b_rate(self, value):
        if self._dir not in ("<>", "<"):
            raise RxDException("no backward reaction in reaction scheme")
        self._original_rate_b = value
        self._update_rates()

    def __repr__(self):
        short_f = (
            self._original_rate_f._short_repr()
            if hasattr(self._original_rate_f, "_short_repr")
            else self._original_rate_f
        )
        short_b = (
            self._original_rate_b._short_repr()
            if hasattr(self._original_rate_b, "_short_repr")
            else self._original_rate_b
        )

        return (
            "MultiCompartmentReaction(%r, %s, rate_b=%s, membrane=%s, custom_dynamics=%r, membrane_flux=%r, scale_by_area=%r)"
            % (
                self._scheme,
                short_f,
                short_b,
                self._regions[0]._short_repr(),
                self._custom_dynamics,
                self._membrane_flux,
                self._scale_by_area,
            )
        )

    def _do_memb_scales(self, cur_map):
        from . import species

        sources = [
            r
            for r in self._sources
            if not isinstance(r(), species.SpeciesOnExtracellular)
        ]
        dests = [
            r
            for r in self._dests
            if not isinstance(r(), species.SpeciesOnExtracellular)
        ]

        # flux occurs on sections which have both source, destination and membrane
        active_secs = list(self._regions[0]._secs1d)
        for sp in sources + dests:
            if sp() and sp()._region():
                active_secs = [
                    sec for sec in active_secs if sec in sp()._region()._secs1d
                ]

        # TODO: Support intracellular 3D reactions
        if not self._scale_by_area:
            narea = sum([sec.nseg for sec in active_secs])
            areas = numpy.ones(narea)
        else:
            # TODO: simplify this expression
            areas = numpy.fromiter(
                itertools.chain.from_iterable(
                    list(
                        self._regions[0]._geometry.volumes1d(sec) for sec in active_secs
                    )
                ),
                dtype=float,
            )
        neuron_areas = []
        for sec in active_secs:
            neuron_areas += [seg.area() for seg in sec]
        neuron_areas = numpy.array(neuron_areas)
        # area_ratios is usually a vector of 1s
        area_ratios = areas / neuron_areas

        # still needs to be multiplied by the valence of each molecule
        self._memb_scales = -area_ratios * h.FARADAY / (10000 * molecules_per_mM_um3())

        # since self._memb_scales is only used to compute currents as seen by the rest of NEURON,
        # we only use NEURON's areas
        # self._memb_scales = volume * molecules_per_mM_um3 / areas

        if not self._membrane_flux:
            return
        # TODO: don't assume/require always inside/outside on one side...
        #       if no nrn_region specified, then (make so that) no contribution
        #       to membrane flux
        sources_ecs = [
            r for r in self._sources if isinstance(r(), species.SpeciesOnExtracellular)
        ]
        dests_ecs = [
            r for r in self._dests if isinstance(r(), species.SpeciesOnExtracellular)
        ]

        source_regions = [s()._region()._nrn_region for s in sources] + [
            "o" for s in sources_ecs
        ]
        dest_regions = [d()._region()._nrn_region for d in dests] + [
            "o" for d in dests_ecs
        ]
        # dereference the species to get the true species if it's actually a SpeciesOnRegion
        sources = [s()._species() for s in self._sources]
        dests = [d()._species() for d in self._dests]

        # TODO: make so don't need multiplicity (just do in one pass)
        smap = {"i": 1, None: 0, "o": -1}
        self._cur_charges = tuple(
            [
                smap[r] * s.charge
                for s, r in zip(sources, source_regions)
                if s.name is not None
            ]
            + [
                -1 * smap[r] * s.charge
                for s, r in zip(dests, dest_regions)
                if s.name is not None
            ]
        )
        self._net_charges = sum(self._cur_charges)
        self._cur_ptrs = []
        self._cur_mapped = []
        self._cur_mapped_ecs = []
        ecs_grids = dict()
        all_grids = []
        for sp in itertools.chain(self._sources, self._dests):
            s = sp()._species()
            if s.name is not None:
                for r in s.regions:
                    if isinstance(s[r], species.SpeciesOnExtracellular):
                        ecs_grids[s.name] = s[r]._extracellular()
                if s.name not in all_grids:
                    all_grids.append((s.name, s.charge))
        for sec in active_secs:
            for seg in sec:
                local_ptrs = []
                local_mapped = []
                local_mapped_ecs = []
                for spname, charge in all_grids:
                    # Check for extracellular regions
                    if charge != 0:
                        name = "_ref_i%s" % (spname)
                        local_ptrs.append(getattr(seg, name))
                        uberlocal_map = [None, None]
                        uberlocal_map_ecs = [None, None]
                        if spname + "i" in cur_map and cur_map[spname + "i"]:
                            uberlocal_map[0] = cur_map[spname + "i"][seg]
                        if spname + "o" in cur_map:
                            # Original rxd extracellular region
                            if seg in cur_map[spname + "o"]:
                                uberlocal_map[1] = cur_map[spname + "o"][seg]
                            elif spname in ecs_grids:  # Extracellular space
                                uberlocal_map_ecs[0] = ecs_grids[
                                    spname
                                ]._grid_id  # TODO: Just pass the grid_id once per species
                                uberlocal_map_ecs[1] = ecs_grids[spname].index_from_xyz(
                                    *species._xyz(seg)
                                )
                        local_mapped.append(uberlocal_map)
                        local_mapped_ecs.append(uberlocal_map_ecs)
                self._cur_ptrs.append(tuple(local_ptrs))
                self._cur_mapped.append(tuple(local_mapped))
                self._cur_mapped_ecs.append(local_mapped_ecs)



"""
UnitTests of the python interface to the hoc.Vector class.

$Id$
"""

import unittest
from neuron import h


class Bench(object):
    def __init__(self, g, l):
        self.g = g
        self.l = l

    def __call__(self, cmd, repeat=5):
        from time import time

        t = 0
        for i in range(repeat):
            t1 = time()
            exec(cmd, self.g, self.l)
            t2 = time()
            t += t2 - t1

        print('Executed "%s".  Elapsed = %f s' % (cmd, t / repeat))


class VectorTestCase(unittest.TestCase):
    """Tests of the hoc.Vector"""

    def testEndian(self):
        """Test that Vector.__array_interface__ returns the proper byteorder (endian)"""

        endian_map = {"little": "<", "big": ">"}

        import sys

        sys_endian = endian_map[sys.byteorder]
        v = h.Vector(10)
        assert sys_endian == v.__array_interface__["typestr"][0]

    def testBytesize(self):
        """Test that Vector.__array_interface__ returns the proper bytesize (of a double)"""

        try:
            import numpy

            v = h.Vector(10)
            a = numpy.array([], dtype=float)
            assert a.__array_interface__["typestr"] == v.__array_interface__["typestr"]
        except:
            pass

    def testPerformance(self):
        """Test performance of Vector<->list,array"""

        try:
            import numpy

            bench = Bench(globals(), locals())
            print("\n")
            bench("l = range(1000000)")
            bench("v = h.Vector(l)")
            print("inplace:")
            bench("v.from_python(l)")
            bench("a = numpy.array(v)")
            print("inplace:")
            bench("v.to_python(a)")
            bench("v2 = h.Vector(a)")
            print("inplace:")
            bench("v2.from_python(a)")
            bench("l2 = list(v2)")
            print("inplace:")
            bench("v.to_python(l2)")
            bench("v2 = h.Vector(a[::-1])")
            bench("a2 = numpy.array(v2)")
        except:
            pass

    def testNumpyInteraction(self):
        """Testing numpy.array <=> hoc.Vector interaction"""

        try:
            import numpy
            from numpy import alltrue, array

            a = numpy.random.normal(size=10000)
            # b = numpy.random.normal(size=10000)
            v = h.Vector(a)
            a1 = array(v)
            assert alltrue(a == a1), 'numpy array "a" not equal to array(Vector(a))'
            v = h.Vector(a[::-1])
            assert alltrue(
                array(v)[::-1] == a
            ), "Vector(a) malfuctions when a is a sliced array"

            # inplace operations

            # todo
        except:
            pass


def suite():

    suite = unittest.makeSuite(VectorTestCase, "test")
    return suite


if __name__ == "__main__":

    # unittest.main()
    runner = unittest.TextTestRunner(verbosity=2)
    runner.run(suite())



import pytest
import gc


@pytest.fixture
def model_rangevarplot(neuron_instance):
    h, rxd, data, save_path = neuron_instance

    def make_rvp(with_sec=True, with_species=True):
        dend = h.Section(name="dend")
        dend.L = 3
        dend.diam = 2
        dend.nseg = 3

        def my_initial(node):
            if h.distance(node, dend(0)) < 2:
                return 1
            else:
                return 0

        cyt = rxd.Region([dend], name="cyt")
        c = rxd.Species(cyt, name="c", initial=my_initial)

        g = h.RangeVarPlot(c, dend(0), dend(1))
        c = c if with_species else None
        dend = dend if with_sec else None
        gc.collect()
        return (dend, cyt, c, g)

    yield (h, rxd, make_rvp)


def test_rangevarplot(model_rangevarplot):
    h, rxd, make_rvp = model_rangevarplot
    dend, cyt, c, g = make_rvp()
    vec = h.Vector()
    g.to_vector(vec)
    assert vec.to_python() == [0, 1, 1, 0, 0]


def test_rangevarplot_no_species(model_rangevarplot):
    h, rxd, make_rvp = model_rangevarplot
    dend, cyt, c, g = make_rvp(with_species=False)
    vec = h.Vector()
    g.to_vector(vec)
    assert vec.to_python() == [0, 0, 0, 0, 0]


def test_rangevarplot_no_section(model_rangevarplot):
    h, rxd, make_rvp = model_rangevarplot
    dend, cyt, c, g = make_rvp(with_sec=False)
    vec = h.Vector()
    g.to_vector(vec)
    for sec in h.allsec():
        assert False
    assert vec.to_python() == [0, 0, 0, 0, 0]



# Sum of all i_membrane_ should equal sum of all ElectrodeCurrent
# For a demanding test, use a tree with many IClamp and ExpSyn point processes
# sprinkled on zero and non-zero area nodes.
import distutils.util
import os

from neuron import h

h.load_file("stdrun.hoc")  # for h.cvode_active


class Cell:
    def __init__(self, id, nsec):
        r = h.Random()
        r.Random123(id, 0, 0)
        nsec += int(r.discunif(0, 4))  # for nontrivial cell_permute=1

        self.id = id
        self.secs = [h.Section(name="d" + str(i), cell=self) for i in range(nsec)]

        # somewhat random tree, d[0] plays role of soma with connections to
        # d[0](0.5) and all others to 1.0
        for i in range(1, nsec):
            iparent = int(r.discunif(0, i - 1))
            x = 0.5 if iparent == 0 else 1.0
            self.secs[i].connect(self.secs[iparent](x))

        # uniform L and diam but somewhat random passive g and e
        for i, sec in enumerate(self.secs):
            sec.L = 10 if i > 0 else 5
            sec.diam = 1 if i > 0 else 5
            sec.insert("pas")
            sec.g_pas = 0.0001 * r.uniform(1.0, 1.1)
            sec.e_pas = -65 * r.uniform(1.0, 1.1)

        # IClamp and ExpSyn at every location (even duplicates) with random
        # parameters (would rather use a Shunt, but ...)
        self.ics = []
        self.syns = []
        self.netcons = []
        self.netstim = h.NetStim()
        self.netstim.number = 1
        self.netstim.start = 0.0
        for sec in self.secs:
            for seg in sec.allseg():
                ic = h.IClamp(seg)
                ic.delay = 0.1
                ic.dur = 1.0
                ic.amp = 0.001 * r.uniform(1.0, 1.1)
                self.ics.append(ic)

                syn = h.ExpSyn(seg)
                syn.e = -65 * r.uniform(1.0, 1.1)
                syn.tau = r.uniform(0.1, 1.0)
                self.syns.append(syn)

                nc = h.NetCon(self.netstim, syn)
                nc.delay = 0.2
                nc.weight[0] = 0.001 * r.uniform(1.0, 1.1)
                self.netcons.append(nc)

    def __str__(self):
        return "Cell" + str(self.id)


def total_imem():
    imem = 0.0
    for sec in h.allsec():
        for seg in sec.allseg():
            if seg.x == 0.0 and sec.parentseg() is not None:
                assert seg.i_membrane_ == sec.parentseg().i_membrane_
                continue  # don't count twice
            imem += seg.i_membrane_
    # print("total_imem ", imem)
    return imem


def total_iclamp(ics):
    icur = 0.0
    for ic in ics:
        icur += ic.i
    # print("total_iclamp ", icur)
    return icur


# verified that events arrived at syn to change g.
def total_syn_g(syns):
    g = 0
    for syn in syns:
        g += syn.g
    # print("total syn g ", g)
    return g


# Remark:
#  The iteration over all segments (including 0 area nodes) without
#  counting 0 area nodes twice is
"""
    for sec in h.allsec():
     for seg in sec.allseg():
       if seg.x == sec.orientation() and sec.parentseg() is not None:
         continue
       # rest of the "for seg..." body
"""
#  Note that sec.parentseg() is needed to count the root and use of
#  sec.trueparentseg() would miss the root node. Also, although an
#  extremely rare edge case, sec.orientation() is needed to match
#  which segment is closest to root.


def test_allseg_unique_iter():
    a = h.Section("a")
    b = h.Section("b")
    c = h.Section("c")
    d = h.Section("d")
    e = h.Section("e")
    f = h.Section("f")
    g = h.Section("g")
    g1 = h.Section("g1")
    g2 = h.Section("g2")
    g3 = h.Section("g3")
    g4 = h.Section("g4")

    b.connect(a(0), 0)
    c.connect(a(0), 0)
    d.connect(a(0), 1)
    e.connect(a(0), 1)
    f.connect(b(1), 0)
    g.connect(b(1), 1)
    g1.connect(g(0), 0)
    g2.connect(g(0), 1)
    g3.connect(g(1), 0)
    g4.connect(g(1), 1)

    # h.topology()

    unique_segs = []
    for sec in h.allsec():
        for seg in sec.allseg():
            if seg.x == sec.orientation() and sec.parentseg() is not None:
                continue
            # print ("%s(%g)" % (sec.name(), seg.x))
            unique_segs.append(seg)

    nseg = sum(sec.nseg for sec in h.allsec())
    nsec = sum(1 for _ in h.allsec())
    ncell = 1
    assert len(unique_segs) == nseg + nsec + ncell
    # nothing left out and nothing done twice
    for sec in h.allsec():
        sec.v = 100.0
    for seg in unique_segs:
        assert seg.v == 100.0  # nothing done twice
        seg.v = 0.0
    for sec in h.allsec():
        for seg in sec.allseg():
            assert seg.v == 0.0  # nothing left out


# verified that there is i_membrane_ != 0.0 in zero area nodes.
def print_imem():
    print("t=%g" % h.t)
    for sec in h.allsec():
        for seg in sec.allseg():
            if seg.x == 0.0 and sec.parentseg() is not None:
                continue  # don't count twice
            print("%s(%g).i_membrane_ = %g" % (sec.name(), seg.x, seg.i_membrane_))


def balanced(ics, tolerance):
    # print_imem() # helped to verify the test is substantive
    bal = abs(total_imem() - total_iclamp(ics))
    if bal > tolerance:
        print(
            "t=%g bal=%g total_imem=%g total_iclamp=%g"
            % (h.t, bal, total_imem(), total_iclamp(ics))
        )
    assert bal <= tolerance


def run(tstop, ics, tolerance):
    # to get nontrivial initialized i_membrane_, initialize to random voltage.
    r = h.Random()
    r.Random123(0, 1, 0)
    for sec in h.allsec():
        for seg in sec.allseg():
            # don't care if some segments counted twice
            seg.v = -65.0 + r.uniform(0, 5)
    h.finitialize()
    balanced(ics, tolerance)
    while h.t < 1.0:
        h.fadvance()
        balanced(ics, tolerance)


def test_fastimem():
    cells = [Cell(id, 10) for id in range(2)]
    # h.topology()
    cvode = h.CVode()
    ics = h.List("IClamp")
    syns = h.List("ExpSyn")
    cvode.use_fast_imem(1)
    h.finitialize(-65)
    run(1.0, ics, 1e-13)
    total_syn_g(syns)
    h.cvode_active(1)
    run(1.0, ics, 1e-12)
    cvode.use_fast_imem(0)
    h.cvode_active(0)


def coreneuron_available():
    if "NRN_ENABLE_CORENEURON=ON" not in h.nrnversion(6):
        # Not ideal. Maybe someday it will be default ON and then
        # will not appear in h.nrnversion(6)
        return False
    # But can it be loaded?
    cvode = h.CVode()
    cvode.cache_efficient(1)
    pc = h.ParallelContext()
    h.finitialize()
    result = 0
    import sys
    from io import StringIO

    original_stderr = sys.stderr
    sys.stderr = StringIO()
    try:
        pc.nrncore_run("--tstop 1 --verbose 0")
        result = 1
    except Exception as e:
        pass
    sys.stderr = original_stderr
    cvode.cache_efficient(0)
    return result


def print_fast_imem():
    ix = h.Vector()
    imem = h.Vector()
    for sec in h.allsec():
        for seg in sec.allseg():
            if seg.x == 0.0 and sec.parentseg() is not None:
                continue  # don't count twice
            ix.append(seg.node_index())
            imem.append(seg.i_membrane_)
    si = ix.sortindex()
    ix.index(ix.c(), si)
    imem.index(imem.c(), si)
    f = open("fastimem.nrn", "w")
    f.write("%d\n" % int(ix.size()))
    for i, x in enumerate(imem):
        assert i == int(ix[i])
        f.write("%d %.20g\n" % (i, x))
    f.close()


def test_fastimem_corenrn():

    if not coreneuron_available():
        return

    print("test_fastimem_corenrn")
    pc = h.ParallelContext()
    ncell = 5
    cvode = h.CVode()
    cvode.cache_efficient(0)
    cells = [Cell(id, 10) for id in range(ncell)]
    cvode.use_fast_imem(1)
    imem = [h.Vector().record(cell.secs[3](0.5)._ref_i_membrane_) for cell in cells]
    tstop = 1.0

    def init_v():
        # to get nontrivial initialized i_membrane_, initialize to random voltage.
        r = h.Random()
        r.Random123(0, 1, 0)
        for sec in h.allsec():
            for seg in sec.allseg():
                # don't care if some segments counted twice
                seg.v = -65.0 + r.uniform(0, 5)
        h.finitialize()

    def run(tstop):
        pc.set_maxstep(10)
        init_v()
        pc.psolve(tstop)

    # standard
    run(tstop)
    imem_std = [vec.c() for vec in imem]

    def compare():
        for i in range(ncell):
            if not imem_std[i].eq(imem[i]):
                print("imem for cell ", i)
                for j, x in enumerate(imem_std[i]):
                    print(j, x, imem[i][j], x - imem[i][j])
            assert imem_std[i].eq(imem[i])
            imem[i].resize(0)

    compare()  # just starting iwth imem cleared

    print("cache efficient NEURON")
    cvode.cache_efficient(1)
    run(tstop)
    compare()

    print("direct mode (online) coreneuron")
    from neuron import coreneuron

    coreneuron.enable = True
    coreneuron.verbose = 0
    coreneuron.gpu = distutils.util.strtobool(
        os.environ.get("CORENRN_ENABLE_GPU", "false")
    )
    run(tstop)
    compare()
    coreneuron.enable = False

    print("Are the i_membrane_ trajectories correct when ...")

    tvec = h.Vector().record(h._ref_t)
    init_v()
    while h.t < tstop - h.dt / 2:
        dt_above = 1.1 * h.dt  # comfortably above dt to avoid 0 step advance
        coreneuron.enable = True
        told = h.t
        pc.psolve(h.t + dt_above)
        assert h.t > told
        coreneuron.enable = False
        pc.psolve(h.t + dt_above)
    compare()

    print("For file mode (offline) coreneuron comparison of i_membrane_ initialization")

    init_v()
    print_fast_imem()

    # The cells must have gids.
    for i, cell in enumerate(cells):
        pc.set_gid2node(i, pc.id())
        sec = cell.secs[0]
        pc.cell(i, h.NetCon(sec(0.5)._ref_v, None, sec=sec))

    # Write the data files
    init_v()
    pc.nrncore_write("./corenrn_data")

    # args needed for offline run of coreneuron
    coreneuron.enable = True
    coreneuron.file_mode = True

    arg = coreneuron.nrncore_arg(tstop)
    coreneuron.enable = False
    pc.gid_clear()
    print(arg)

    cvode.use_fast_imem(0)


if __name__ == "__main__":
    test_allseg_unique_iter()
    test_fastimem()
    test_fastimem_corenrn()
    h.quit()



from neuron import h, gui
from neuron.expect_hocerr import expect_hocerr

pc = h.ParallelContext()


def run(tstop):
    h.tstop = tstop
    h.run()


def on_finit(ste, result):
    result.clear()
    ste.state(0)


def act(i, m, thresh, result):
    print(
        "act %d t=%g (%g) v=%g thresh=%g" % (i, h.t, pc.t(0), m["s"](0.5).v, thresh[0])
    )
    # allow verification of result
    result.append((i, thresh[0], m["vvec"].size() - 1))


def chk_result(sz, result, m):
    print(result)
    for x in result:
        th = x[1]
        i = x[2]
        vvec = m["vvec"]
        print(
            "vvec[%d]=%.18g th=%.18g vvec[%d]=%.18g"
            % (i - 1, vvec[i - 1], th, i, vvec[i])
        )
        assert vvec[i] >= th - h.float_epsilon and vvec[i - 1] <= th
    assert len(result) == sz


def model():
    # model with linear increase in voltage
    s = h.Section()
    s.L = 10
    s.diam = 10
    ic = h.IClamp(s(0.5))
    ic.delay = 0
    ic.dur = 10
    ic.amp = 0.1
    vvec = h.Vector().record(s(0.5)._ref_v, sec=s)
    tvec = h.Vector().record(h._ref_t, sec=s)
    return {"s": s, "ic": ic, "vvec": vvec, "tvec": tvec}


def test_ste():

    m1 = model()
    # one state ste with two self transitions
    var = m1["s"](0.5)._ref_v
    thresh1 = h.ref(-50)
    thresh2 = h.ref(-10)
    result = []
    ste = h.StateTransitionEvent(1)
    ste.transition(0, 0, var, thresh1, (act, (1, m1, thresh1, result)))
    ste.transition(0, 0, var, thresh2, (act, (2, m1, thresh2, result)))
    fih = h.FInitializeHandler((on_finit, (ste, result)))

    run(5)
    print("final v=%g" % m1["s"](0.5).v)
    chk_result(2, result, m1)

    h.cvode_active(1)
    run(20)
    chk_result(2, result, m1)

    h.cvode_active(1)
    h.cvode.condition_order(2)
    run(5)
    chk_result(2, result, m1)

    h.cvode.condition_order(1)
    h.cvode_active(0)

    # ste associated with point process
    del fih, ste
    ste = h.StateTransitionEvent(2, m1["ic"])
    fih = h.FInitializeHandler((on_finit, (ste, result)))
    run(5)

    # transition with hoc callback
    h("""proc foo() { printf("foo called at t=%g\\n", t) }""")
    thresh3 = h.ref(-30)
    ste.transition(0, 0, var, thresh3, "foo()")
    run(5)

    # transition with hoc callback in hoc object
    h(
        """
begintemplate FooSTEtest
objref this
proc foo() { printf("foo in %s called at t=%g\\n", this, t) }
endtemplate FooSTEtest
"""
    )
    thresh4 = h.ref(-20)
    obj = h.FooSTEtest()
    ste.transition(0, 0, var, thresh4, "foo()", obj)
    run(5)

    del ste, fih
    assert h.List("StateTransitionEvent").count() == 0
    assert h.List("FInitializeHandler").count() == 0


if __name__ == "__main__":
    test_ste()



def psection(sec):
    from neuron import h

    try:
        from neuron import rxd
        from neuron.rxd import region, species
        from neuron.rxd import rxd as rxd_module

        have_rxd = True
    except:
        have_rxd = False

    results = {}

    mname = h.ref("")

    # point processes
    pps = {}
    mt = h.MechanismType(1)
    for i in range(int(mt.count())):
        mt.select(i)
        mypps = set()
        pp = mt.pp_begin(sec=sec)
        while pp is not None:
            mypps.add(pp)
            pp = mt.pp_next()
        if mypps:
            mt.selected(mname)
            pps[mname[0]] = mypps
    results["point_processes"] = pps

    center_seg_dir = dir(sec(0.5))

    mechs_present = []

    # membrane mechanisms
    mt = h.MechanismType(0)

    for i in range(int(mt.count())):
        mt.select(i)
        mt.selected(mname)
        name = mname[0]
        if name in center_seg_dir:
            mechs_present.append(name)

    results["density_mechs"] = {}
    results["ions"] = {}

    for mech in mechs_present:
        my_results = {}
        ms = h.MechanismStandard(mech, 0)
        for j in range(int(ms.count())):
            n = int(ms.name(mname, j))
            name = mname[0]
            pvals = []
            # TODO: technically this is assuming everything that ends with _ion
            #       is an ion. Check this.
            if mech.endswith("_ion"):
                pvals = [getattr(seg, name) for seg in sec]
            else:
                mechname = name  # + '_' + mech
                for seg in sec:
                    if n > 1:
                        pvals.append([getattr(seg, mechname)[i] for i in range(n)])
                    else:
                        pvals.append(getattr(seg, mechname))
            my_results[
                name[: -(len(mech) + 1)] if name.endswith("_" + mech) else name
            ] = pvals
        # TODO: should be a better way of testing if an ion
        if mech.endswith("_ion"):
            results["ions"][mech[:-4]] = my_results
        else:
            results["density_mechs"][mech] = my_results

    morphology = {
        "L": sec.L,
        "diam": [seg.diam for seg in sec],
        "pts3d": [
            (sec.x3d(i), sec.y3d(i), sec.z3d(i), sec.diam3d(i))
            for i in range(sec.n3d())
        ],
        "parent": sec.parentseg(),
        "trueparent": sec.trueparentseg(),
    }

    results["morphology"] = morphology
    results["nseg"] = sec.nseg
    results["Ra"] = sec.Ra
    results["cm"] = [seg.cm for seg in sec]

    if have_rxd:
        regions = {
            r() for r in region._all_regions if r() is not None and sec in r().secs
        }
        results["regions"] = regions

        my_species = []
        for sp in species._all_species:
            sp = sp()
            if sp is not None:
                sp_regions = sp._regions
                if not hasattr(sp_regions, "__len__"):
                    sp_regions = [sp_regions]
                if any(r in sp_regions for r in regions):
                    my_species.append(sp)
        results["species"] = set(my_species)
        results["name"] = sec.hname()
        results["hoc_internal_name"] = sec.hoc_internal_name()
        results["cell"] = sec.cell()

        # all_active_reactions = [r() for r in rxd_module._all_reactions if r() is not None]

    return results


if __name__ == "__main__":
    from pprint import pprint
    from neuron import h, rxd

    soma = h.Section(name="soma")
    soma.insert("hh")
    soma.insert("pas")
    soma.nseg = 3
    iclamp = h.IClamp(soma(0.3))
    gaba = h.ExpSyn(soma(0.7))
    dend = h.Section(name="dend")
    dend.connect(soma)
    cyt = rxd.Region([dend], name="cyt", nrn_region="i")
    er = rxd.Region([dend], name="er")
    na = rxd.Species(cyt, name="na", charge=1)
    ca = rxd.Species([cyt, er], name="ca", charge=2)
    h.finitialize(-65)
    pprint(psection(soma))
    print("\n")
    pprint(psection(dend))



from neuron import h
import subprocess
from subprocess import PIPE

pc = h.ParallelContext()
cvode = h.CVode()


def sortspikes(spiketime, gidvec):
    return sorted(zip(spiketime, gidvec))


# Passive cell random tree so there is some inhomogeneity of ia and im
class Cell:
    def __init__(self, id, nsec):
        r = h.Random()
        r.Random123(id, 0, 0)
        nsec += int(r.discunif(0, 4))  # for nontrivial cell_permute=1

        self.id = id
        self.secs = [h.Section(name="d" + str(i), cell=self) for i in range(nsec)]

        # somewhat random tree, d0 plays role of soma with child connections to
        # d0(.5) and all others to  1.0
        for i in range(1, nsec):
            iparent = int(r.discunif(0, i - 1))
            x = 0.5 if iparent == 0 else 1.0
            self.secs[i].connect(self.secs[iparent](x))

        # uniform L and diam but somewhat random passive g and e
        for i, sec in enumerate(self.secs):
            sec.nseg = int(r.discunif(3, 5)) if i > 0 else 1
            sec.L = 200 if i > 0 else 5
            sec.diam = 1 if i > 0 else 5
            sec.insert("pas")
            sec.g_pas = 0.0001 * r.uniform(1.0, 1.1)
            sec.e_pas = -65 * r.uniform(1.0, 1.1)
        self.secs[0].insert("hh")

        # IClamp at d2(0)
        # As i_membane_ does not include ELECTRODE_CURRENT, comparison with
        # im_axial should be done after ic completes.
        ic = h.IClamp(self.secs[2](0))
        ic.delay = 0.0
        ic.dur = 0.2
        ic.amp = 5.0
        self.ic = ic

        # axial at every internal segment, a unique AxialPP at root
        # and at every sec(1) zero area node.
        self.axialpps = [h.AxialPP(self.secs[0](0))]
        for sec in self.secs:
            sec.insert("axial")
            self.axialpps.append(h.AxialPP(sec(1)))
        # update pointers and ri later after all cells are created

    def get_axial(self, seg):
        if 0.0 < seg.x < 1.0:
            mech = seg.axial
        else:
            mech = None
            pps = seg.point_processes()
            for pp in pps:
                if type(pp) == type(self.axialpps[0]):
                    mech = pp
                    break
        assert mech is not None
        return mech

    def __str__(self):
        return "Cell" + str(self.id)

    def ptr_helper(self, seg, mech, pseg, pmech):
        # mech and pmech for both axial and AxialPP have same rangevars.
        if pmech is None:
            mech.ri = 0.0  # indicates pointers not to be used
            # POINTER unused but must not be NULL for CoreNEURON so point to self
            mech._ref_pv = seg._ref_v
            mech._ref_pia = mech._ref_ia
            mech._ref_pim = mech._ref_im
        else:
            mech.ri = seg.ri()
            mech._ref_pv = pseg._ref_v
            mech._ref_pia = pmech._ref_ia
            mech._ref_pim = pmech._ref_im

    def update_pointers(self):
        # Probably works if every section has default orientation, i.e.
        # arc position 0 at proximal (toward the root) end.
        for i, sec in enumerate(self.secs):
            seg = sec(0)
            pseg = sec.trueparentseg()  # None only for root
            pmech = None
            if i == 0:
                assert pseg is None
                mech = self.axialpps[0]
                self.ptr_helper(seg, mech, pseg, pmech)
            else:
                # Already dealt with proximal end of section at some previous
                # iteration. But need to figure out mech
                assert pseg is not None
                mech = self.get_axial(pseg)

            pseg = seg
            pmech = mech
            for seg in sec:
                mech = seg.axial
                self.ptr_helper(seg, mech, pseg, pmech)
                pseg = seg
                pmech = mech

            # distal end of section. pseg and pmech are correct
            seg = sec(1)
            mech = self.axialpps[
                i + 1
            ]  # 0 is for rootnode, 1 is distal of root section
            self.ptr_helper(seg, mech, pseg, pmech)


class Model:
    def __init__(self, ncell, nsec):
        self.cells = [Cell(i, nsec) for i in range(ncell)]
        self.update_pointers()
        # Setup callback to update dipole POINTER for cache_efficiency
        # The PtrVector is used only to support the callback.
        self._callback_setup = h.PtrVector(1)
        self._callback_setup.ptr_update_callback(self.update_pointers)

    def update_pointers(self):
        for cell in self.cells:
            cell.update_pointers()


def srun(cmd):
    print("--------------------")
    print(cmd)
    r = subprocess.run(cmd, shell=True, stdout=PIPE, stderr=PIPE)
    if r.returncode != 0:
        print(r)
    r.check_returncode()


def runcn(args):
    import platform

    cpu = platform.machine()
    cmd = cpu + "/special-core " + args
    srun(cmd)


def test_axial():
    m = Model(5, 5)
    cvode.use_fast_imem(1)

    def result(m):
        v = []
        ia = []
        im = []
        imem = []
        loc = []

        def add(cell, seg):  # for axial and AxialPP
            mech = cell.get_axial(seg)
            loc.append((seg, mech, mech.im))
            v.append(seg.v)
            ia.append(mech.ia)
            im.append(mech.im)
            imem.append(seg.i_membrane_)

        for cell in m.cells:
            add(cell, cell.secs[0](0))
            for sec in cell.secs:
                for seg in sec:
                    add(cell, seg)
                add(cell, sec(1))

        # im == imem
        mx = max([abs(y) for y in imem])
        mx = mx if mx > 0 else 1.0
        x = sum([abs(imem[i] - y) for i, y in enumerate(im)])
        assert x / mx < 1e-9
        return v, ia, im, imem, loc

    def chk(std, result):
        # cell_permute = 1 --- im and imem addition is not associative
        # assert std == result
        for i in [0, 1, 2, 3]:
            mx = max([abs(x) for x in std[i]])
            mx = mx if mx > 0 else 1.0
            x = sum([abs(std[i][j] - x) for j, x in enumerate(result[i])])
            assert x / mx < 1e-9

    def run(tstop):
        pc.set_maxstep(10)
        h.finitialize(-65)
        pc.psolve(tstop)
        return result(m)

    std = run(1)

    cvode.cache_efficient(1)
    chk(std, run(1))

    from neuron import coreneuron

    coreneuron.verbose = 0
    coreneuron.enable = True
    coreneuron.cell_permute = 0
    for coreneuron.cell_permute in [0, 1]:
        chk(std, run(1))
    coreneuron.enable = False

    m._callback_setup = None  # get rid of the callback first.
    del m


def test_checkpoint():
    if pc.nhost() > 1:
        return

    # clear out the old
    srun("rm -r -f coredat")

    m = Model(5, 5)
    # file mode CoreNEURON real cells need gids
    for i, cell in enumerate(m.cells):
        pc.set_gid2node(i, pc.id())
        sec = cell.secs[0]
        pc.cell(i, h.NetCon(sec(0.5)._ref_v, None, sec=sec))

    # "integrate" fabs(ia) in each axial_pp and use that as a source of spikes
    # for actually testing that the checkpoint is working with POINTER.
    # Would be much better if I knew how to get file mode coreneuron to
    # print trajectories.
    for i, p in enumerate(h.List("AxialPP")):
        pc.set_gid2node(i + 100, pc.id())
        pc.cell(i + 100, h.NetCon(p, None))

    spktime = h.Vector()
    spkgid = h.Vector()
    pc.spike_record(-1, spktime, spkgid)
    cvode.cache_efficient(1)
    pc.set_maxstep(10)
    h.finitialize(-65)
    pc.nrncore_write("coredat")

    # do a NEURON run to record spikes
    def run(tstop):
        pc.set_maxstep(10)
        h.finitialize(-65)
        pc.psolve(tstop)

    run(10)
    spikes_std = sortspikes(spktime, spkgid)

    # Does it work in direct mode?
    from neuron import coreneuron

    coreneuron.enable = True
    for perm in [0, 1]:
        coreneuron.cell_permute = perm
        run(5)
        pc.psolve(10)
        spikes = sortspikes(spktime, spkgid)
        assert spikes_std == spikes
    coreneuron.enable = False

    # standard to compare with checkpoint series
    tpnts = [5.0, 10.0]
    for perm in [0, 1]:
        print("\n\ncell_permute ", perm)
        common = "-d coredat --voltage 1000 --verbose 0 --cell-permute %d" % (perm,)
        # standard full run
        runcn(common + " --tstop %g" % float(tpnts[-1]) + " -o coredat")
        # sequence of checkpoints
        for i, tpnt in enumerate(tpnts):
            tend = tpnt
            restore = " --restore coredat/chkpnt%d" % (i,) if i > 0 else ""
            checkpoint = " --checkpoint coredat/chkpnt%d" % (i + 1,)
            outpath = " -o coredat/chkpnt%d" % (i + 1,)
            runcn(
                common + " --tstop %g" % (float(tend),) + outpath + restore + checkpoint
            )

        # compare spikes
        cmp_spks(
            spikes_std, "coredat", ["coredat/chkpnt%d" % (i,) for i in range(1, 3)]
        )

    m._callback_setup = None
    pc.gid_clear()
    del m


def cmp_spks(spikes, dir, chkpntdirs):
    # sorted nrn standard spikes into dir/out.spk
    f = open(dir + "/temp", "w")
    for spike in spikes:
        f.write("%.8g\t%d\n" % (spike[0], int(spike[1])))
    f.close()
    # sometimes roundoff to %.8g gives different sort.
    srun("sortspike {}/temp {}/nrn.spk".format(dir, dir))

    srun("sortspike {}/out.dat {}/out.spk".format(dir, dir))
    srun("cmp {}/out.spk {}/nrn.spk".format(dir, dir))

    cmd = "cat"
    for i in chkpntdirs:
        cmd = cmd + " " + i + "/out.dat"
    srun(cmd + " > " + dir + "/temp")
    srun("sortspike {}/temp {}/chkptout.spk".format(dir, dir))
    srun("cmp {}/out.spk {}/chkptout.spk".format(dir, dir))


if __name__ == "__main__":
    test_axial()
    test_checkpoint()



"""reads and plots data from a file compatible with
stdgui.hoc's clipboard_retrieve()
"""

from neuron import h, gui

h.clipboard_retrieve()  # user selects file to be read, results in h.hoc_obj_[1] and [0]
xvec = h.hoc_obj_[1]
yvec = h.hoc_obj_[0]

if xvec.size() <= 1:
    print("insufficient data")
    h.quit()

# now xvec and yvec contain the xy coords of the points

# up to this point, everything has been generic
# below this point, everything is customized by the user

from plotfi import plotfi

# creates a Graph with customized label and axes
fig = plotfi(xvec, yvec)



"""
Some tests of nrnpy_pyCallObject that need to be run manually as they involve
errors generated via gui usage. Run with 'python -i gui_pycallobject.py'
axonerr: ZeroDivisionError: division by zero
axonexit: exits immediately (need 'stty sane' in termina)
In the Graph, select 'mouse events'.
  hold Crtl key and press mouse button: ZeroDivisionError
  hold Shift key and press mouse button: exits immediately (need 'stty sane')

"""

from neuron import h, gui
import sys

soma = h.Section(name="soma")
axon = h.Section(name="axon")
axonerr = h.Section(name="axonerr")
axonexit = h.Section(name="axonexit")


def select(sec):
    print("select: {} {}".format(sec, type(sec)))
    if sec == axonerr:
        print("substantial memory leak, e.g. ivHit, ivEvent")
        print(1 / 0)
    if sec == axonexit:
        print('need "stty sane" on exit')
        sys.exit(0)


def accept(sec):
    print("accept: {}".format(sec))


sb = h.SectionBrowser()
sb.select_action(select)
sb.accept_action(accept)

my_list = h.List()
for word in ["NEURON", "HOC", "Python", "NMODL"]:
    my_list.append(h.String(word))


def label_with_lengths():
    item_id = int(h.hoc_ac_)
    if item_id == 4:
        1 / 0
    if item_id == 5:
        sys.exit(0)
    item = my_list[item_id].s
    return "%s (%d)" % (item, len(item))


my_list.browser("Words!", label_with_lengths)
# first append will generate error but will add item to list
try:
    my_list.append(h.String("error"))
except:
    print("appending 'error' failed")

# second append will exit
my_list.append(h.String("exit"))
print("this does not print if not iteractive.")
assert my_list.count() == 6


def on_event(event_type, x, y, keystate):
    print("{} {} {} {}".format(event_type, x, y, keystate))
    if keystate == 1:
        1 / 0
    if keystate == 2:
        sys.exit(0)


g = h.Graph()
g.menu_tool("mouse events", on_event)



from neuron import h

pc = h.ParallelContext()
rank = pc.id()
nhost = pc.nhost()
pc.nthread(4)

# Want to exercise the internal indexing schemes. So need mpi and threads.
# Random order of sgid calls to source_var. Random order and location
# of sgid source . Random order and location of sgid target. An sgid source
# has random number of targets.
# Might be a good idea to add some voltage sources and ki sources.

ncell = 100
nsrc = 50  # so that number of sgid pointing to _ref_nai
# multiple sgid may NOT point to same _ref_nai
ntarget = 200  # so sgid on average sends to ntarget/nsrc targets.

h(
    """
begintemplate Cell
public soma
create soma
proc init() {
  soma.diam = 5
  soma.L = 5
  soma {
    insert na_ion
    ion_style("na_ion", 1,0,0,0,0) // finitialize leaves nai as is
      // Otherwise cannot pass interpreter setting onto coreneuron
      // unless use a mod file that WRITE nai in INITIAL via parameter
  }
}
endtemplate Cell
"""
)


def test_natrans():
    gids = [gid for gid in range(rank, ncell, nhost)]
    cells = []
    for gid in range(rank, ncell, nhost):
        pc.set_gid2node(gid, rank)
        cells.append(h.Cell())
        pc.cell(gid, h.NetCon(cells[-1].soma(0.5)._ref_v, None, sec=cells[-1].soma))

    r = h.Random()
    r.Random123(1, 1, 0)

    # nsrc unique random sgids in range(ncell). Sample without replacement.
    sgids = []
    v = list(range(ncell))
    for i in range(nsrc):
        x = int(r.discunif(0, len(v) - 1))
        sgids.append(v[x])
        v.pop(x)
    del v

    for sgid in sgids:
        if pc.gid_exists(sgid) == 3:
            sec = pc.gid2cell(sgid).soma
            pc.source_var(sec(0.5)._ref_nai, sgid, sec=sec)

    # ntarget randomly chosen cells are targets for the nsrc sgids
    # ntarget NaTrans are created
    targets = []
    for itar in range(ntarget):
        gid = int(r.discunif(0, ncell - 1))
        sgid = sgids[int(r.discunif(0, nsrc - 1))]
        if pc.gid_exists(gid) == 3:
            sec = pc.gid2cell(gid).soma
            target = h.NaTrans(sec(0.5))
            targets.append(target)
            pc.target_var(target, target._ref_napre, sgid)
            target.sgid = sgid

    cvode = h.CVode()
    cvode.cache_efficient(1)

    pc.set_maxstep(10)
    pc.setup_transfer()

    h.dt = 0.1
    tstop = 0.1

    def run():
        h.finitialize(-65)
        for sgid in sgids:
            if pc.gid_exists(sgid) == 3:
                sec = pc.gid2cell(sgid).soma
                sec(0.5).nai = float(sgid) / 100.0 + 0.001
        tars = h.List("NaTrans")
        for tar in tars:
            tar.napre = 0.0001  # correct values don't carryover from previous sim
        pc.psolve(tstop)
        for tar in tars:
            x = (tar.sgid / 100.0 + 0.001) if tar.sgid >= 0.0 else 0.0
            differ = ("differ") if abs(tar.napre - x) > 1e-10 else ""
            if differ != "":
                print(
                    "%d %s %g %g %g %s"
                    % (rank, tar.hname(), tar.sgid, tar.napre, x, differ)
                )
            assert differ == ""

    run()  # NEURON: Fails if tar.napre not what is expected

    from neuron import coreneuron

    coreneuron.available = True
    if coreneuron.available:
        coreneuron.enable = True
        coreneuron.cell_permute = 0
        run()  # Fails if CoreNEURON does not copy expected tar.napre to NEURON

    return cells, gids, sgids, targets


if __name__ == "__main__":
    model = test_natrans()
    pc.barrier()
    h.quit()



import pytest
from .testutils import compare_data, tol


def test_parameter_reaction(neuron_instance):
    """Test a simple rxd.Reaction involving a rxd.Parameter"""
    h, rxd, data, save_path = neuron_instance
    sec = h.Section(name="sec")
    cyt = rxd.Region([sec], name="cyt", nrn_region="i")
    ca = rxd.Species(cyt, name="ca", charge=2, initial=0.001)
    buf = rxd.Parameter(cyt, name="buf", value=0.001)
    cabuf = rxd.Species(cyt, name="cabuf", initial=0)
    buffering = rxd.Reaction(ca + buf, cabuf, 0.02, 0.01)
    buffering2 = rxd.Reaction(ca + buf, cabuf, 0.0, 0.0)
    badrate = rxd.Rate(buf, -buf)
    h.finitialize(-65)
    h.continuerun(10)
    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
import subprocess

# Make translators & domains available for include
sys.path.insert(0, os.path.abspath("./translators"))
sys.path.insert(0, os.path.abspath("./domains"))

import html2  # responsible for creating jump tables in python and hoc documentation
import hocdomain  # Sphinx HOC domain (hacked from the Python domain via docs/generate_hocdomain.py)

# -- Project information -----------------------------------------------------

project = "NEURON"
copyright = "2022, Duke, Yale and the Blue Brain Project"
author = "Michael Hines"

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.autosectionlabel",
    "recommonmark",
    "sphinx.ext.mathjax",
    "nbsphinx",
    "sphinx_design",
]

source_suffix = {
    ".rst": "restructuredtext",
    ".txt": "markdown",
    ".md": "markdown",
}


def setup(app):
    """Setup connect events to the sitemap builder"""
    app.set_translator("html", html2.HTMLTranslator)

    # Set-up HOC domain
    hocdomain.setup(app)


# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "python/venv"]

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Sphinx expects the master doc to be contents
master_doc = "index"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Extra html content that is generated. i.e. doxygen
html_extra_path = ["_generated"]

html_css_files = [
    "custom.css",
]

# We never execute the notebooks via nbsphinx (for faster local iterations)
# notebooks are executed thanks to the `notebooks` CMake target
nbsphinx_execute = "never"

# Force mathjax@v2 due to plotly requirement
# https://www.sphinx-doc.org/en/master/usage/extensions/math.html#module-sphinx.ext.mathjax
mathjax_path = (
    "https://cdn.jsdelivr.net/npm/mathjax@2/MathJax.js?config=TeX-AMS-MML_HTMLorMML"
)
mathjax2_config = {
    "tex2jax": {
        "inlineMath": [["$", "$"], ["\\(", "\\)"]],
        "processEscapes": True,
        "ignoreClass": "document",
        "processClass": "math|output_area",
    }
}

if os.environ.get("READTHEDOCS"):
    # Get RTD build version ('latest' for master and actual version for tags)
    # Use alias PKGVER to avoid mixin' with sphinx and wasting lots of time on debugging that
    from packaging import version as PKGVER

    rtd_ver = PKGVER.parse(os.environ.get("READTHEDOCS_VERSION"))

    # Install neuron accordingly (nightly for master, otherwise incoming version)
    # Note that neuron wheel must be published a priori.
    subprocess.run(
        "pip install neuron{}".format(
            "=={}".format(rtd_ver.base_version)
            if isinstance(rtd_ver, PKGVER.Version)
            else "-nightly"
        ),
        shell=True,
        check=True,
    )

    # Execute & convert notebooks + doxygen
    subprocess.run("cd .. && python setup.py docs", check=True, shell=True)



from neuron import h

h.load_file("stdrun.hoc")


class Cell(object):
    def __init__(self):
        self.topology()
        self.subsets()
        self.geometry()
        self.biophys()
        self.synapses()

    def topology(self):
        self.soma = h.Section(cell=self)
        self.dend = h.Section(cell=self)
        self.dend.connect(self.soma)
        self.nseg = 5

    def subsets(self):
        self.all = h.SectionList()
        self.all.wholetree(sec=self.soma)

    def geometry(self):
        self.soma.L = 10
        self.soma.diam = 10
        self.dend.L = 500
        self.dend.diam = 1

    def biophys(self):
        for sec in self.all:
            sec.Ra = 100
            sec.cm = 1
        self.soma.insert("hh")

    def synapses(self):
        self.syn = h.ExpSyn(0.5, sec=self.dend)
        self.syn.e = 0
        self.syn.tau = 1

    def connectToTarget(self, syn):
        nc = h.NetCon(self.soma(0.5)._ref_v, syn, sec=self.soma)
        nc.threshold = -10
        return nc


# read which group we are part of
# in the old c interface it was
# g = new_intp()
# if MUSIC_configInt(s, 'group', g):
#  group = intp_value(g)
#  delete_intp(g)
# else:
#  group = 0
# del(g)
from sys import argv
from getopt import gnu_getopt

optlist, args = gnu_getopt(argv[3:], "g:")
group = int(optlist[0][1])

pc = h.ParallelContext()
rank = int(pc.id())
print("group=", group, " rank=", rank)

gid = group
pc.set_gid2node(gid, pc.id())
cell = Cell()
pc.cell(gid, cell.connectToTarget(None))
if gid == 1:
    stim = h.IClamp(cell.soma(0.5))
    stim.delay = 0
    stim.dur = 0.5
    stim.amp = 0.2


from neuronmusic import *

out = publishEventOutput("output")
inp = publishEventInput("input")

out.gid2index(gid, gid - 1)
nc = inp.index2target(gid % 2, cell.syn)
nc.weight[0] = 0.005
nc.delay = 8

pc.set_maxstep(5)
print(group, rank, " call stdinit")
h.stdinit()
print(group, rank, " call psolve")
pc.psolve(50)
print(group, rank, " done")



cdef extern from "music.hh":
    ctypedef struct cxx_Setup "MUSIC::Setup"

cdef extern from "nrnmusic.h":
    ctypedef struct cxx_EventOutputPort "NRNMUSIC::EventOutputPort":
       void gid2index (int gid, int gi)

    cxx_EventOutputPort* new_EventOutputPort "new NRNMUSIC::EventOutputPort" (cxx_Setup* s, char* id)
    void del_EventOutputPort "delete" (cxx_EventOutputPort* obj)

    ctypedef struct cxx_EventInputPort "NRNMUSIC::EventInputPort":
       object index2target (int gi, object target)

    cxx_EventInputPort* new_EventInputPort "new NRNMUSIC::EventInputPort" (cxx_Setup* s, char* id)
    void del_EventInputPort "delete" (cxx_EventInputPort* obj)

    cxx_EventOutputPort* cxx_publishEventOutput "NRNMUSIC::publishEventOutput" (char* identifier)

    cxx_EventInputPort* cxx_publishEventInput "NRNMUSIC::publishEventInput" (char* identifier)


# Local Variables:
# mode: python
# End:



from neuron import h
from neuron.units import ms, mV
import matplotlib.pyplot as plt
from ring import Ring

ring = Ring()

pc = h.ParallelContext()
pc.set_maxstep(10 * ms)

t = h.Vector().record(h._ref_t)
h.finitialize(-65 * mV)
pc.psolve(100 * ms)

# send all spike time data to node 0
local_data = {cell._gid: list(cell.spike_times) for cell in ring.cells}
all_data = pc.py_alltoall([local_data] + [None] * (pc.nhost() - 1))

if pc.id() == 0:
    # combine the data from the various processes
    data = {}
    for process_data in all_data:
        data.update(process_data)
    # plot it
    plt.figure()
    for i, spike_times in data.items():
        plt.vlines(spike_times, i + 0.5, i + 1.5)
    plt.show()

pc.barrier()
pc.done()
h.quit()



import pytest
from .testutils import compare_data, tol


@pytest.fixture
def model_pump(neuron_instance):
    """A simple rxd implementation of the sodium-potassium pump"""

    h, rxd, data, save_path = neuron_instance
    dend = h.Section(name="dend")
    dend.diam = 2
    dend.nseg = 101
    dend.L = 100

    cyt = rxd.Region(dend, name="cyt", nrn_region="i")
    ecs = rxd.Region(dend, name="ecs", nrn_region="o")
    mem = rxd.Region(dend, name="mem", geometry=rxd.membrane())
    na = rxd.Species(
        [cyt, ecs],
        name="na",
        charge=1,
        initial=lambda nd: 18 if nd.region == cyt else 144,
    )
    k = rxd.Species(
        [cyt, ecs],
        name="k",
        charge=1,
        initial=lambda nd: 140 if nd.region == cyt else 3,
    )

    nai, nao, ki, ko = na[cyt], na[ecs], k[cyt], k[ecs]
    exp = rxd.rxdmath.exp
    pump = (0.8 / (1.0 + exp((25.0 - nai) / 3.0))) * (1.0 / (1.0 + exp(3.5 - ko)))
    volume_scale = 1e-18 * rxd.constants.NA() * (dend.diam / 4.0)
    pump_current = rxd.MultiCompartmentReaction(
        2 * ko + 3 * nai,
        2 * ki + 3 * nao,
        pump * volume_scale,
        mass_action=False,
        membrane=mem,
        membrane_flux=True,
    )
    model = (
        dend,
        cyt,
        ecs,
        mem,
        na,
        k,
        nai,
        nao,
        ki,
        ko,
        pump,
        volume_scale,
        pump_current,
    )
    yield (neuron_instance, model)


def test_currents(model_pump):
    """Test currents generated by a Na/K-pump fixed step integration."""

    neuron_instance, model = model_pump
    h, rxd, data, save_path = neuron_instance
    # check changing the units after initialization
    h.nrnunit_use_legacy(False)
    h.finitialize(-65)
    h.nrnunit_use_legacy(True)
    h.continuerun(10)
    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol


def test_currents_cvode(model_pump):
    """Test currents generated by a Na/K-pump variable step integration."""

    neuron_instance, model = model_pump
    h, rxd, data, save_path = neuron_instance
    # check changing the units after initialization
    h.CVode().active(True)
    h.nrnunit_use_legacy(False)
    h.finitialize(-65)
    h.nrnunit_use_legacy(True)
    h.continuerun(10)
    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol



# More testing  BBSaveState.

from neuron import h

pc = h.ParallelContext()
cvode = h.CVode()


class Model1:
    def __init__(self, ncell, n):
        # gid -> [n NetCons] -> gid+1
        cells = {gid: h.Follower() for gid in range(pc.id(), ncell, pc.nhost())}
        for gid in cells:
            pc.set_gid2node(gid, pc.id())
            pc.cell(gid, h.NetCon(cells[gid], None))
        netcons = {}
        for gid in cells:
            netcons[gid] = [
                pc.gid_connect(gid - 1, cells[gid]) for _ in range(n) if gid > 0
            ]
        # Stimulate gid 0
        if 0 in cells:
            # stim cell[0] with fast burst of 5 spikes
            ns = h.NetStim()
            ns.start = 0.9999
            ns.number = 5
            ns.interval = 0.1
            nsnc = h.NetCon(ns, cells[0])
            nsnc.delay = 0
            nsnc.weight[0] = 0
        for gid, cell in cells.items():
            # cells generate spike on every input of weight > 1
            # cell.refrac = 0
            for i, nc in enumerate(netcons[gid]):
                nc.weight[0] = 0.001 * i + h.dt
                nc.delay = (1.0 + 0.01 * i) * h.dt

        self.cells = cells
        self.netcons = netcons
        self.ns = ns
        self.nsnc = nsnc


class Model2:
    """NetStim -> Cell -> N Cells -> Cell"""

    # exercise PreSyn on queue with use_min_delay_ by means of
    # several NetCon with same PreSyn and same delay
    def __init__(self, n):
        self.ncell_per_layer = [1, n, 2]
        self.cells = {}
        # make cells and associate with gid (round robin distribution)
        for ilayer, ncell in enumerate(self.ncell_per_layer):
            for icell in range(ncell):
                gid = self.info2gid(ilayer, icell)
                if (gid % pc.nhost()) == pc.id():
                    self.cells[gid] = h.Follower()
                    pc.set_gid2node(gid, pc.id())
                    pc.cell(gid, h.NetCon(self.cells[gid], None))

        # make connections (all to all from layer i-1 to layer i)
        self.netcons = {}
        for gid, cell in self.cells.items():
            ilayer, icell = self.gid2info(gid)
            if ilayer == 0:
                continue
            srclayer = ilayer - 1
            for isrc in range(self.ncell_per_layer[srclayer]):
                srcgid = self.info2gid(srclayer, isrc)
                nc = pc.gid_connect(srcgid, cell)
                nc.weight[0] = 0.1
                nc.delay = 1.0
                self.netcons[(srcgid, gid)] = nc

        # Stimulate layer 0 with NetStim
        self.netstims = {}
        self.netstim_con = {}
        for gid in range(self.ncell_per_layer[0]):
            if gid in self.cells:
                ns = h.NetStim()
                self.netstims[gid] = ns
                ns.start = 0.9999
                ns.number = 5
                ns.interval = 0.1
                nc = h.NetCon(ns, self.cells[gid])
                nc.delay = 0.0
                nc.weight[0] = 0.1
                self.netstim_con[gid] = nc
        self.ns = self.netstims[0]
        self.nsnc = self.netstim_con[0]

        print(self.cells)
        print(self.netcons)
        print(self.netstim_con)

    def info2gid(self, ilayer, icell):
        return ilayer * 100 + icell

    def gid2info(self, gid):
        return (int(gid / 100)), gid % 100


def test_bbss():
    print("focus on BinQ initialization Issue #1444")
    ncell = 3
    n = 5

    model = Model2(5)

    #    pc.gid_clear()
    #    model = Model1(ncell, n)

    spiketime = h.Vector()
    spikegid = h.Vector()
    pc.spike_record(-1, spiketime, spikegid)

    def run(tstop):
        pc.set_maxstep(10)
        h.finitialize()
        pc.psolve(tstop)

    spiketime_std = spiketime.c()
    spikegid_std = spikegid.c()

    def set_stdspikes():
        spiketime_std.copy(spiketime)
        spikegid_std.copy(spikegid)

    tstop = 5
    run(tstop)
    set_stdspikes()

    def prspikes():
        print("prspikes")
        for i, gid in enumerate(spikegid):
            print("%g %d" % (spiketime[i], gid))

    def compare_spikes():
        x = list(zip(spiketime_std, spikegid_std))
        x = sorted(x, key=lambda e: e[1])
        y = list(zip(spiketime, spikegid))
        y = sorted(y, key=lambda e: e[1])
        if len(x) != len(y):
            print(len(x), len(y))
        # assert len(x) == len(y)
        if x != y:
            q = (x, y) if len(x) <= len(y) else (y, x)
            for i, a in enumerate(q[0]):
                b = q[1][i]
                if a != b:
                    # assert a[1] == b[1]
                    z = abs(a[0] - b[0])
                    if z >= 1e-9:
                        print(x[i], y[i])
                    assert z < 1e-9

    def srun(tsave, tstop):
        qm = cvode.queue_mode()
        print(
            "srun tsave=%g tstop=%g start=%g delay=%g %s"
            % (tsave, tstop, model.ns.start, model.nsnc.delay, "binq" if qm else "")
        )
        run(tsave)
        st = spiketime.c()
        sg = spikegid.c()
        bbss = h.BBSaveState()
        bbss.save("allcell_bbss.dat")
        qm = 1
        if qm:
            print("after save")
            cvode.print_event_queue()
        h.finitialize(-65)
        spiketime.resize(0)
        spikegid.resize(0)
        bbss.restore("allcell_bbss.dat")
        if qm:
            print("after restore")
            cvode.print_event_queue()
        pc.psolve(tstop)
        st.append(spiketime)
        sg.append(spikegid)
        spiketime.copy(st)
        spikegid.copy(sg)
        print("    %d spikes" % len(spiketime))

    srun(1.1, tstop)
    compare_spikes()

    cvode.queue_mode(1)
    run(tstop)
    assert len(spiketime_std) == len(spiketime)
    spiketime_std = spiketime.c()
    spikegid_std = spikegid.c()

    srun(1.1, tstop)
    compare_spikes()
    for binq in [0, 1]:
        cvode.queue_mode(binq)
        parms = [
            (1, 0),
            (1e-10, 0),
            (0, 0),
            (0.25 * h.dt, 0),
            (0.5 * h.dt, 0),
            (0.75 * h.dt, 0),
        ]
        for parm in parms:
            if 0 in model.cells:
                model.ns.start, model.nsnc.delay = parm
            for tsave in [0.0, h.dt]:
                run(tstop)
                set_stdspikes()
                srun(tsave, tstop)
                compare_spikes()

    cvode.queue_mode(0)
    pc.gid_clear()


if __name__ == "__main__":
    test_bbss()



from neuron import h, crxd as rxd

h.load_file("stdrun.hoc")
soma = h.Section(name="soma")
r = rxd.Region([soma])
soma.nseg = 5
c = rxd.Species(r, initial=1.0)
assert len(c.nodes) == 5
soma.nseg = 3
assert len(c.nodes) == 3
soma.nseg = 7
assert len(c[r].nodes) == 7
h.tstop = 100
h.run()



from neuron import h, rxd

h.load_file("stdrun.hoc")


class Cell:
    def __init__(self):
        self.soma = h.Section()


cell_1 = Cell()
cell_2 = Cell()
ecs = rxd.Extracellular(
    -55, -55, -55, 55, 55, 55, dx=10, volume_fraction=0.2, tortuosity=1.6
)

soma_cell_1 = rxd.Region([cell_1.soma], nrn_region="i", geometry=rxd.Shell(0.8, 1))
soma_cell_2 = rxd.Region([cell_2.soma], nrn_region="i", geometry=rxd.Shell(0.8, 1))
species_x = rxd.Species(
    [ecs, soma_cell_1, soma_cell_2],
    name="species_x",
    d=1,
    charge=1,
    initial=1,
    ecs_boundary_conditions=0,
)
Species_y = rxd.Species(
    [soma_cell_1, ecs],
    name="species_y",
    d=1,
    charge=1,
    initial=1,
    ecs_boundary_conditions=0,
)
Species_z = rxd.Species(
    [soma_cell_1], name="species_z", d=1, charge=1, initial=1, ecs_boundary_conditions=0
)
reaction_1 = rxd.Reaction(
    Species_y > Species_z + species_x, 0.058, regions=[soma_cell_1]
)

h.finitialize()
h.continuerun(10)



#
# See http://neuroml.svn.sourceforge.net/viewvc/neuroml/NeuroML2nrn/
# for the latest version of this file.
#

# 1 for verbose debug mode
debug = 0


class Point:
    def __init__(self, id, parent_id, cable_id, lineno):
        self.id_ = id
        self.pid_ = parent_id
        self.cid_ = cable_id
        self.lineno_ = lineno

    def set(self, x, y, z, diam):
        self.x_ = x
        self.y_ = y
        self.z_ = z
        self.d_ = diam

    def __str__(self):
        return (
            "Point: "
            + str(self.id_)
            + ", ("
            + str(self.x_)
            + ", "
            + str(self.y_)
            + ", "
            + str(self.z_)
            + " | "
            + str(self.d_)
            + ") , par: "
            + str(self.pid_)
            + ", cable id: "
            + str(self.cid_)
        )


class Cable:
    def __init__(self, id, pid, ipnt):
        self.first_ = ipnt
        self.pcnt_ = 0
        self.id_ = id
        self.pid_ = pid
        self.px_ = -1.0
        self.parent_cable_id_ = -1
        self.name_ = ""
        if debug:
            print(" New ", self.__str__())

    def __str__(self):
        return (
            "Cable: id="
            + str(self.id_)
            + " name="
            + self.name_
            + " pid="
            + str(self.parent_cable_id_)
            + " first="
            + str(self.first_)
        )


class CableGroup:
    def __init__(self, name):
        self.name_ = name
        self.cable_indices_ = []
        self.mechs_ = []

    def __str__(self):
        info = "CableGroup: " + self.name_
        for ci in self.cable_indices_:
            info = info + "\n" + "  " + str(ci)
        return info


class BioParm:
    def __init__(self, name, value):
        self.name_ = name
        self.value_ = value
        self.group_index_ = -1

    def __str__(self):
        return (
            "BioParm: "
            + self.name_
            + ", val: "
            + str(self.value_)
            + ", grp id: "
            + str(self.group_index_)
        )


class BioMech:
    def __init__(self, name):
        self.name_ = name
        self.parms_ = []

    def __str__(self):
        info = "BioMech: " + self.name_
        for p in self.parms_:
            info = info + "\n" + "  " + str(p)
        return info


class LineNo:
    def __init__(self):
        self.lineno = 0

    def getLineNumber(self):
        return self.lineno


class XML2Nrn:
    def __init__(self):
        if debug:
            print("Starting parsing of NeuroML file... ")
        self.in_biogroup_ = False
        self.in_paramgroup_ = False
        self.locator = LineNo()

    def prattrs(self, node):
        for i in list(node.keys()):
            print("  ", i, " ", node.get(i))



from neuron import h, gui

soma = h.Section(name="soma")
soma.insert(h.hh)
soma.L = 5
soma.diam = 5

iclamp = h.IClamp(soma(0.5))
iclamp.delay = 0.1
iclamp.dur = 0.1
iclamp.amp = 0.3

g = h.Graph(False)
g.view(0, -80, 5, 120, 624, 40, 300, 200)
h.graphList[0].append(g)
g.addvar("v", soma(0.5)._ref_v)

# popup the variable step and run controls
h.numericalmethodpanel.map()
h.nrncontrolmenu()



from neuron import h
from neuron.expect_hocerr import expect_hocerr, expect_err, set_quiet
import sys

quiet = True

uni = "ab\xe0"


def checking(s):
    if not quiet:
        print("CHECKING: " + s)


class Foo:  # for testing section name errors when full cell.section name has unicode chars.
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return self.name


def test_py2nrnstring():
    print(uni)

    # I don't think h.getstr() can be made to work from python
    # so can't test unicode on stdin read by hoc.
    h("strdef s")
    h('s = "hello"')
    checking("h('getstr(s)')")
    h("getstr(s)")
    "goodbye"
    assert h.s == "hello"

    checking("h(uni + ' = 1')")
    assert h(uni + " = 1") == 0

    # Allowed! The s format for PyArg_ParseTuple specifies that
    # Unicode objects are converted to C strings using 'utf-8' encoding.
    # If this conversion fails, a UnicodeError is raised.
    checking("""h('s = "%s"'%uni)""")
    assert h('s = "%s"' % uni) == 1
    assert h.s == uni

    checking('h.printf("%s", uni)')
    expect_hocerr(h.printf, ("%s", uni))

    checking("hasattr(h, uni)")
    expect_hocerr(hasattr, (h, uni))

    expect_err("h.à = 1")

    expect_err("a = h.à")

    expect_err('a = h.ref("à")')

    ns = h.NetStim()
    # nonsense but it does test the unicode error message
    checking('h.setpointer(ns._ref_start, "à", ns)')
    expect_hocerr(h.setpointer, (ns._ref_start, "à", ns))
    del ns

    # No error for these two (unless cell is involved)!
    checking("s = h.Section(name=uni)")
    s = h.Section(name=uni)
    assert s.name() == uni
    checking("s = h.Section(uni)")
    s = h.Section(uni)
    assert s.name() == uni

    expect_err('h.Section(name="apical", cell=Foo(uni))')

    soma = h.Section()
    expect_err("a = soma.à")

    expect_err("soma.à = 1")

    expect_err("a = soma(.5).à")

    expect_err("soma(.5).à = 1")

    soma.insert("hh")
    expect_err("a = soma(.5).hh.à")

    expect_err("soma(.5).hh.à = 1")


if __name__ == "__main__":
    set_quiet(False)
    quiet = False
    test_py2nrnstring()



import distutils.util
import os

from neuron import h

pc = h.ParallelContext()


def test_units():
    s = h.Section()
    pp = h.UnitsTest(s(0.5))
    h.ion_style("na_ion", 1, 2, 1, 1, 0, sec=s)

    h.finitialize(-65)
    R_std = pp.gasconst
    erev_std = pp.erev
    ghk_std = pp.ghk

    from neuron import coreneuron

    h.CVode().cache_efficient(1)
    coreneuron.enable = True
    coreneuron.gpu = bool(
        distutils.util.strtobool(os.environ.get("CORENRN_ENABLE_GPU", "false"))
    )
    pc.set_maxstep(10)
    h.finitialize(-65)
    pc.psolve(h.dt)

    assert R_std == pp.gasconst  # mod2c needs nrnunits.lib.in
    assert abs(erev_std - pp.erev) <= (
        1e-13 if coreneuron.gpu else 0
    )  # GPU has tiny numerical differences
    assert ghk_std == pp.ghk


if __name__ == "__main__":
    test_units()
    h.quit()



"""
UnitTests of the python interface to the neuron class.
Items declared in neuron/__init__.py

$Id$
"""

import unittest
import neuron
from neuron import h


class NeuronTestCase(unittest.TestCase):
    """Tests of neuron"""

    def testHClass(self):
        """Test subclass of hoc class."""

        from ._subclass import A1

        a = A1(5)
        assert a.x == 5.0
        assert a.p() == 6.0
        b = A1(4)
        a.s = "one"
        b.s = "two"
        assert a.s == "one"
        assert b.s == "two"
        assert h.A[0].s == "one"
        assert a.p() == 7.0
        assert b.p() == 5.0
        a.a = 2
        b.a = 3
        assert a.a == 2
        assert b.a == 3
        assert h.List("A").count() == 2
        a = 1
        b = 1
        assert h.List("A").count() == 0

    @classmethod
    def psection(cls):
        """Test neuron.psection(Section)"""

        s = h.Section(name="soma")
        neuron.psection(s)

    def testpsection(self):
        from multiprocessing import Process

        p = Process(target=NeuronTestCase.psection)
        p.start()
        p.join()

    def testABI(self):
        """Test use of some  Py_LIMITED_API for python3."""

        # Py_nb_bool
        assert True if h else False
        assert True if h.List else False
        # ensure creating a List doesn't change the truth value
        l = h.List()
        assert True if h.List else False
        assert False if l else True
        v = h.Vector(1)
        l.append(v)
        assert True if l else False

        # Py_sq_length
        assert len(l) == 1
        # Py_sq_item
        assert l[0] == v
        # Py_sq_ass_item
        v.x[0] = 5
        assert v.x[0] == 5

    def testIterators(self):
        """Test section, segment, mechanism, rangevar iterators."""
        # setup model
        sections = [h.Section(name="s%d" % i) for i in range(3)]
        iclamps = [h.IClamp(sec(0.5)) for sec in sections]
        for i, sec in enumerate(sections):
            sec.nseg = 3
            sec.insert("pas")
            sec.insert("hh")
        # iterate
        import hashlib

        sha = hashlib.sha256()
        for sec in h.allsec():
            for seg in sec:
                for mech in seg:
                    for var in mech:
                        txt = "%s(%g).%s.%s=%g" % (
                            sec.name(),
                            seg.x,
                            mech.name(),
                            var.name(),
                            var[0],
                        )
                        sha.update(txt.encode("utf-8"))
        d = sha.hexdigest()
        d1 = "ac49344c054bc9e56e165fa75423d8bcb7cce96c4527f259362b527ee05103d8"
        # in case NRN_ENABLE_MOD_COMPATIBILITY=ON
        # (set by -DNRN_ENABLE_CORENEURON=ON)
        d2 = "44366906aa94a50644bc734eb23afcc25d1206c0431c4e7908698eeb2597c385"
        assert d == d1 or d == d2
        sections[0](0.5).na_ion.ena = 40.0  # issue #651
        assert sections[0](0.5).na_ion.ena == 40.0

    def testSectionArgOrder(self):
        """First optional arg for Section is name (but name="name" is recommended)"""
        soma = h.Section("soma")
        assert soma.name() == "soma"

    def testSectionCell(self):
        """Section.cell() internally referenced as weakref."""
        err = -1
        try:
            soma = h.Section(cell="foo", name="soma")
            err = 1
        except:
            err = 0
        assert err == 0

        class Cell:
            def __str__(self):
                return "hello"

        c = Cell()
        soma = h.Section(cell=c, name="soma")
        assert soma.name() == "hello.soma"
        assert soma.cell() == c
        del c
        assert soma.cell() is None

    def testSectionListIterator(self):
        """As of v8.0, iteration over a SectionList does not change the cas"""
        # See issue 509. SectionList iterator bug requires change to
        # longstanding behavior
        soma = h.Section(name="soma")
        soma.push()
        sections = [h.Section(name="s%d" % i) for i in range(3)]
        assert len([s for s in h.allsec()]) == 4
        sl = h.SectionList(sections)
        # Iteration over s SectionList does not change the currently accessed section
        for s in sl:
            assert 1 and h.cas() == soma
        # If an iteration does not complete the section stack is still ok.
        assert sections[1] in sl
        assert 2 and h.cas() == soma

    @classmethod
    def ExtendedSection(cls):
        """test prsection (modified print statement)"""
        from neuron.sections import ExtendedSection

        s = ExtendedSection(name="test")
        s.psection()

    def testExtendedSection(self):
        from multiprocessing import Process

        p = Process(target=NeuronTestCase.ExtendedSection)
        p.start()
        p.join()

    @classmethod
    def RxDexistence(cls):
        """test import rxd and geometry3d"""
        error = 0
        try:
            from neuron import rxd
            from neuron.rxd import geometry

            print("has_geometry3d is " + str(geometry.has_geometry3d))
        except Exception as e:
            print("'from neuron import rxd' failed", e)
            error = 1
        else:
            try:
                a = basicRxD3D()
                print("    basicRxD3D() ran with no exception")
            except Exception as e:
                print("'basicRxD3D()' failed", e)
                error = 1
        assert error == 0
        return 0

    def testHelp(self):
        error = False
        try:
            from neuron import doc

            print(doc.get_docstring("xpanel", ""))
        except Exception as e:
            print("'doc.get_docstring('xpanel', '')' failed:", e)
            error = True
        self.assertFalse(error)
        return 0

    def testRxDexistence(self):
        from multiprocessing import Process

        p = Process(target=NeuronTestCase.RxDexistence)
        p.start()
        p.join()
        assert p.exitcode == 0
        return 0

    def test_newobj_err(self):
        """Test deletion of incompletely constructed objects"""
        print()  # Error message not on above line
        h.load_file("stdlib.hoc")  # need hoc String
        h(
            """
begintemplate Foo
endtemplate Foo

begintemplate NewObj
objref this, ob, foo1, foo2
proc init() {localobj s
  foo1 = new Foo() // Constructed before error, even partial constructions fill this field.
  if ($1 == 0) {
    execerror("generate an error") // All NewObj instances undergoing construction
  } else if ($1 == $2) {
    // This and all NewObj instances prior to this will construct successfully.
    // All after this will be partially constructed.
    // The execerror should cause only the partially constructed NewObj to
    // be destroyed.
    s = new String()
    sprint(s.s, "ob = new NewObj(%d, %d)", $1-1, $2)
    execute1(s.s, this)
  } else {
    ob = new NewObj($1-1, $2)
  }
  foo2 = new Foo() // Only instances prior to execute1 reach here.
}
endtemplate NewObj
"""
        )
        # arg[0] recursion depth
        # arg[0] - arg[1] + 1 should be successfully constructed
        # arg[1] should be partially constructed and destroyed.
        args = (4, 2)
        a = h.NewObj(*args)
        b = h.List("NewObj")
        c = h.List("Foo")
        print("#NewObj and #Foo in existence", b.count(), c.count())
        z = args[0] - args[1] + 1
        assert b.count() == z
        assert c.count() == 2 * z

        del a
        del b
        del c
        b = h.List("NewObj")
        c = h.List("Foo")
        print("after del a #NewObj and #Foo in existence", b.count(), c.count())
        assert b.count() == 0
        assert c.count() == 0

        return 1


def basicRxD3D():
    from neuron import h, rxd

    s = h.Section(name="s")
    s.L = s.diam = 1
    cyt = rxd.Region([s])
    ca = rxd.Species(cyt)
    rxd.set_solve_type(dimension=3)
    h.finitialize(-65)
    h.fadvance()
    return 1


def suite():

    suite = unittest.makeSuite(NeuronTestCase, "test")
    return suite


if __name__ == "__main__":

    # unittest.main()
    runner = unittest.TextTestRunner(verbosity=2)
    runner.run(suite())



from math import pi

import pytest

from testutils import compare_data, tol


@pytest.fixture
def ics_example(neuron_instance):
    """A model to test 3D intracellular rxd is correctly handling NMODL
    currents.
    """

    h, rxd, data, save_path = neuron_instance
    rxd.set_solve_type(dimension=3)
    # create cell1 where `x` will be created and leak out
    cell1 = h.Section(name="cell1")
    cell1.pt3dclear()
    cell1.pt3dadd(-1, 0, 0, 1)
    cell1.pt3dadd(1, 0, 0, 1)
    cell1.nseg = 11
    cell1.insert("pump")

    # Where?
    # the intracellular spaces
    cyt = rxd.Region(h.allsec(), name="cyt", nrn_region="i", dx=1.0)

    model = (cell1, cyt)
    yield (neuron_instance, model)


def test_ics_currents(ics_example):
    """Test ics_example with fixed step methods"""

    (h, rxd, data, save_path), (cell1, cyt) = ics_example
    x = rxd.Species([cyt], name="x", d=1.0, charge=1, initial=0)
    h.finitialize(-65)
    h.continuerun(100)

    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol


def test_ics_currents_cvode(ics_example):
    """Test ics_example with variable step methods"""

    (h, rxd, data, save_path), (cell1, cyt) = ics_example
    x = rxd.Species([cyt], name="x", d=1.0, charge=1, initial=0)
    h.CVode().active(True)

    h.finitialize(-65)
    h.continuerun(100)

    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol


def test_ics_currents_initial(ics_example):
    """Test ics_example with initial values from NEURON"""

    (h, rxd, data, save_path), (cell1, cyt) = ics_example
    x = rxd.Species([cyt], name="x", d=1.0, charge=1)
    h.finitialize(-65)
    h.continuerun(100)

    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol



from neuron import h
from neuron.expect_hocerr import expect_err
from checkresult import Chk

import io, math, os, re, sys

dir_path = os.path.dirname(os.path.realpath(__file__))
chk = Chk(os.path.join(dir_path, "test_netcvode.json"))
if hasattr(h, "usetable_hh"):
    h.usetable_hh = 0  # So same whether compiled with CoreNEURON or not.

cv = h.CVode()
pc = h.ParallelContext()

# remove address info from cv.debug_event output
def debug_event_filter(s):
    s = re.sub(r"cvode_0x[0-9abcdef]* ", "cvode_0x... ", s)
    s = re.sub(r"Vector\[\d+\]", "Vector[N]", s)
    return s


# helpful for debugging when expect no NetCon exist
def prNetCon(s):
    print(s)
    ncl = h.List("NetCon")
    print("NetCon count ", len(ncl))
    for nc in ncl:
        print("    ", nc, nc.pre(), nc.preseg(), nc.syn())


# helpful for debugging when expect no Sections
def prSection(s):
    print(s)
    nsec = sum(1 for _ in h.allsec())
    print("Section count ", nsec)
    for sec in h.allsec():
        print("    ", sec.name())


# call netcvode.cpp: void nrn_use_daspk(int b)
def nrn_use_daspk():
    cv.use_daspk(0)
    s = h.Section()
    cv.active(1)
    s.insert("extracellular")
    h.finitialize(-65)
    s.diam = 200
    h.finitialize(-65)
    assert cv.use_daspk() == 1.0
    del s
    cv.use_daspk(0)
    cv.active(0)


# call netcvode: static Node*(Object*)
def node():
    # Requires source POINT_PROCESS with "x" threshold variable and lvardt
    # Such a POINT_PROCESS is discouraged in favor of NET_RECEIVE with WATCH.
    s = h.Section()
    s.L = 10
    s.diam = 10
    cv.active(1)
    ic = h.IClamp(s(0.5))
    ic.delay = 0
    ic.dur = 5
    ic.amp = 0.3
    src = h.PPxSrc(s(0.5))
    nc = h.NetCon(src, None)

    results = [
        [
            (0.6500000000999997, 12.070427805842268),
            (0.7000315963318277, 16.84809332604498),
            (0.7000315963318277, 16.84809332604498),
        ],
        [
            (0.6500000000999997, 12.070427805842268),
            (0.6283185307179584, 10.0),
            (0.6283185307179584, 10.0),
        ],
    ]

    def ev(*arg):
        print("ev t=%g v=%g x=%g nc.x=%g" % (h.t, s(0.5).v, src.x, nc.x))
        ref_t, ref_x = results[arg[0]][arg[1]]
        assert h.t == ref_t
        assert math.isclose(src.x, ref_x, rel_tol=1e-13)

    def run():
        order = 0
        h.dt = 0.025
        type = (
            "fixed" if cv.active() == 0 else "lvardt" if cv.use_local_dt() else "cvode"
        )
        method = 0 if cv.active() == 0 else 2 if cv.use_local_dt() else 1
        if cv.active():
            type += " condition_order %d" % cv.condition_order()
            order = 0 if cv.condition_order() == 1 else 1
        nc.record((ev, (order, method)))
        print("\n" + type, "  thresh ", nc.threshold)
        pc.set_maxstep(10)
        h.finitialize(-50)
        pc.psolve(1)

    def series(condorder):
        cv.condition_order(condorder)
        cv.active(0)
        run()
        cv.active(1)
        run()
        cv.use_local_dt(1)
        run()
        cv.use_local_dt(0)
        cv.active(0)

    # otherwise, the first order threshold crossing is missed
    cv.maxstep(0.1)
    series(1)
    cv.maxstep(1000)
    series(2)
    del nc


def p(s):
    print(s)
    ncl = h.List("NetCon")
    print("NetCon count ", len(ncl))
    for nc in ncl:
        print("    ", nc, nc.pre(), nc.preseg(), nc.syn())


class Cell:
    def __init__(self, id):
        self.id = id
        self.soma = h.Section(name="soma", cell=self)
        self.soma.diam = 10.0
        self.soma.L = 10.0
        self.soma.insert("hh")

    def __str__(self):
        return "Cell" + str(self.id)


h(
    """ // HOC cells work best for CVode.netconlist
begintemplate HCell
public soma
create soma
proc init() {
  create soma
  soma {
    diam = 10
    L = 10
    insert hh
  }
}
endtemplate HCell
"""
)


class Net:
    # all to all
    def __init__(self, ncell, hcell=False):
        cells = (
            [Cell(i) for i in range(ncell)]
            if hcell == False
            else [h.HCell() for _ in range(ncell)]
        )
        syns = {}  # {(i,j):(ExpSyn,NetCon)}
        for i in range(ncell):
            for j in range(ncell):
                syn = h.ExpSyn(cells[j].soma(0.5))
                # spikes via inhibitory rebound and synchronizes
                syn.e = -120
                syn.tau = 3
                nc = h.NetCon(cells[i].soma(0.5)._ref_v, syn, sec=cells[i].soma)
                nc.delay = 5
                nc.weight[0] = 0.001
                nc.threshold = -10
                syns[(i, j)] = (syn, nc)
        ic = h.IClamp(cells[0].soma(0.5))
        ic.delay = 0
        ic.dur = 0.1
        ic.amp = 0.3
        self.cells = cells
        self.syns = syns
        self.ic = ic


def netcon_access():
    net = Net(4)

    # NetCon.preloc and NetCon.preseg
    x = net.syns[(0, 1)][1].preloc()
    sec = h.cas()
    h.pop_section()
    assert sec(x) == net.cells[0].soma(0.5)
    assert net.syns[(0, 1)][1].preseg() == net.cells[0].soma(0.5)
    # two more lines of coverage if source is at 0 loc of section
    nc = h.NetCon(
        net.cells[0].soma(0)._ref_v, net.syns[(0, 1)][0], sec=net.cells[0].soma
    )
    x = nc.preloc()
    sec = h.cas()
    h.pop_section()
    assert sec(x) == net.cells[0].soma(0)
    assert nc.preseg() == net.cells[0].soma(0)
    del nc
    # if NetCon source not a section then return -1 or None
    acell = h.NetStim()
    nc = h.NetCon(acell, net.syns[(0, 1)][0])
    x = nc.preloc()
    assert x == -1.0  # and no need to pop the section
    assert nc.preseg() == None
    del nc, acell
    # if NetCon source not a voltage but in a section then return -2 or None
    nc = h.NetCon(
        net.cells[0].soma(0.5).hh._ref_m, net.syns[(0, 1)][0], sec=net.cells[0].soma
    )
    x = nc.preloc()
    assert x == -2.0
    h.pop_section()
    assert nc.preseg() == None
    del nc

    # NetCon.postloc, NetCon.postseg, NetCon.syn, NetCon.pre
    x = net.syns[(0, 1)][1].postloc()
    sec = h.cas()
    h.pop_section()
    assert sec(x) == net.cells[1].soma(0.5)
    assert net.syns[(0, 1)][1].postseg() == net.cells[1].soma(0.5)
    assert net.syns[(0, 1)][1].pre() == None
    # if NetCon.target not associated with a section then return -1 or None
    a = [h.IntFire1() for _ in range(2)]
    nc = h.NetCon(a[0], a[1])
    x = nc.postloc()
    assert x == -1.0
    assert nc.postseg() == None
    assert nc.pre() == a[0]
    assert nc.syn() == a[1]
    nc = h.NetCon(a[0], None)
    assert nc.syn() == None
    del nc, a, x, sec

    # NetCon.synlist, NetCon.prelist, NetCon.postcelllist, NetCon.precelllist
    # NetCon.precell, NetCon.postcell NetCon.wcnt
    a = [h.IntFire1() for _ in range(2)]
    nclist = [h.NetCon(a[0], a[1]) for _ in range(10)]
    nclist2 = nclist[0].synlist()
    assert nclist2.count() == 10
    nclist2 = h.List()
    nclist[0].synlist(nclist2)
    assert nclist2.count() == 10
    nclist2 = nclist[0].prelist()
    assert nclist2.count() == 10
    net = Net(4, True)  # til fix nrn_sec2cell memory leak
    nclist2 = net.syns[(0, 1)][1].prelist()
    assert nclist2.count() == len(net.cells)
    nclist2 = net.syns[(0, 1)][1].precelllist()
    assert nclist2.count() == len(net.cells)
    nclist2 = net.syns[(0, 1)][1].postcelllist()
    assert nclist2.count() == len(net.cells)
    nclist2 = nclist[0].postcelllist()  # works only for Real cells
    assert nclist2.count() == 0
    assert nclist[0].precell() == None
    assert nclist[0].postcell() == None
    assert net.syns[(2, 3)][1].precell() == net.cells[2]
    assert net.syns[(2, 3)][1].postcell() == net.cells[3]
    del nclist, nclist2, a
    # NetCon.setpost
    net.syns[(0, 1)][1].setpost(None)
    assert net.syns[(0, 1)][1].postseg() == None
    expect_err("net.syns[(0, 1)][1].setpost([])")
    net.syns[(0, 1)][1].setpost(net.syns[(0, 1)][0])
    assert net.syns[(0, 1)][1].syn() == net.syns[(0, 1)][0]
    # added a setpost test in coreneuron/test_fornetcon.py to cover the
    # case of a change in weight vector size.

    # NetCon.valid NetCon.active
    net = Net(2)
    assert net.syns[(0, 1)][1].active()
    assert net.syns[(0, 1)][1].active(0) == True
    assert net.syns[(0, 1)][1].active() == False
    assert net.syns[(0, 1)][1].active(1) == False
    assert net.syns[(0, 1)][1].valid()
    net.syns[(0, 1)][1].setpost(None)
    assert net.syns[(0, 1)][1].valid() == False

    # NetCon.event
    net = Net(2)
    net.ic.amp = 0.0  # no intrinsic activity
    pc.set_maxstep(10)
    h.finitialize(-65)
    net.syns[(0, 1)][1].active(0)
    assert net.syns[(0, 1)][1].event(0.1) == 0
    net.syns[(0, 1)][1].active(1)
    expect_err("net.syns[(0, 1)][1].event(0.1, 1)")
    del net
    cv.active(0)
    h.dt = 0.025
    a = h.IntFire1()
    nc = h.NetCon(None, a)
    h.finitialize()
    nc.weight[0] = 2
    nc.event(0)  # when delivered a fires and turns on refractory
    h.fadvance()
    assert a.m == 2.0
    nc.weight[0] = 0.2
    nc.event(h.t, 1)  # when delivered turns off refractory
    h.fadvance()
    assert a.m == 0
    nc.event(h.t)  # so when delivered this event changes m
    h.fadvance()
    assert a.m == 0.2

    # NetCon.record cover a line, NetCon.get_recordvec, NetCon.wcnt
    assert nc.wcnt() == 1
    tv = h.Vector()
    nc.record(tv)
    assert nc.get_recordvec() == tv
    nc.record()
    nc = h.NetCon(None, None)
    expect_err("nc.record()")  # cover static void NetCon::chksrc()

    # NetCon.srcgid cover a line
    nc = h.NetCon(a, a)
    del a
    assert nc.srcgid() == -1
    nc = h.NetCon(None, None)
    assert nc.srcgid() == -1
    del nc

    # NetCon constructor errors and some coverage
    expect_err("h.NetCon([], None)")
    expect_err("h.NetCon(None, [])")
    h.NetCon(None, None, -10.0, 2.0, 0.001)
    a = h.IntFire1()
    h.NetCon(a, a, -10.0, 2.0, 0.001)
    net = Net(2)
    h.NetCon(
        net.cells[0].soma(0.5)._ref_v, None, -10.0, 2.0, 0.001, sec=net.cells[0].soma
    )
    del net, a

    # CVode.netconlist
    net = Net(2, True)  # Hoc Cell test til nrn_sec2cell memory leak fixed
    a = h.IntFire1()
    b = h.NetStim()
    nc = h.NetCon(b, a)
    lst = cv.netconlist("", "", "")
    assert len(h.List("NetCon")) == 5
    assert len(cv.netconlist(b, "", "")) == 1
    assert len(cv.netconlist("NetStim[<0-9>*]", "", "")) == 1
    assert len(cv.netconlist(b, "", a)) == 1
    assert len(cv.netconlist(b, "", "IntFire1")) == 1
    assert len(cv.netconlist(net.cells[0], "", "")) == 2
    assert len(cv.netconlist(net.cells[0].hname(), "", "")) == 2
    assert len(cv.netconlist("", net.cells[0], "")) == 2
    assert len(cv.netconlist("", net.cells[0].hname(), "")) == 2
    assert len(cv.netconlist("", "", net.syns[0, 1][0])) == 1
    assert len(cv.netconlist("", "", net.syns[0, 1][0].hname())) == 1

    expect_err('cv.netconlist("foo<<", "", "")')
    expect_err('cv.netconlist("", "foo<<",  "")')
    expect_err('cv.netconlist("", "", "foo<<")')
    del lst, nc, a, b, net
    locals()


def cvode_meth():
    # empty
    s = h.Vector()
    ds = h.Vector()
    cv.states(s)
    assert len(s) == 0
    cv.dstates(ds)
    assert len(ds) == 0
    h.finitialize(-65)
    cv.event(0.5)
    cv.solve(1.0)
    cv.event(1.1)
    cv.solve()

    cv.active(1)
    pc.nthread(2)
    h.finitialize(-65)
    cv.event(0.5)
    cv.solve(1)
    cv.event(1.01)
    cv.solve()
    pc.nthread(1)
    cv.use_local_dt(1)
    assert cv.current_method() == 300.0
    cv.use_local_dt(0)

    net = Net(2)
    cv.active(1)
    cv.use_local_dt(0)
    h.finitialize(-65)
    cv.solve(1.0)
    cv.statistics()
    cv.spike_stat(ds)
    # Note: states, dstates, acor are not the same for lvardt
    cv.states(s)
    sref = h.ref("")
    snames = [(cv.statename(i, sref), sref[0])[1] for i in range(len(s))]
    print(snames)
    chk("cv.statename", snames)
    chk("cv.states", s, tol=1e-8)
    cv.dstates(ds)
    chk("cv.dstates", ds, tol=1e-8)
    vec = h.Vector()
    cv.error_weights(vec)
    chk("cv.error_weights", vec)
    cv.acor(vec)
    chk("cv.acor", vec, tol=1e-7)
    std = (h.t, s.to_python(), ds.to_python())
    ds.fill(0)
    cv.f(1.0, s, ds)
    print("std ", std)
    print("(t, s, ds)", (h.t, s.to_python(), ds.to_python()))
    assert (h.t, s.to_python(), ds.to_python()) == std
    s.resize(0)
    expect_err("cv.f(1.0, s, ds)")
    pc.nthread(2)
    h.finitialize(-65)
    s = h.Vector(std[1])
    expect_err("cv.f(std[0], s, ds)")
    pc.nthread(1)
    h.finitialize(-65)

    cv.use_local_dt(1)
    h.finitialize(-65)
    cv.solve(1.0)
    cv.statistics()
    cv.spike_stat(ds)
    expect_err("cv.f(std[0], s, ds)")
    cv.states(s)
    sref = h.ref("")
    snames = [(cv.statename(i, sref), sref[0])[1] for i in range(len(s))]
    chk("cv.statename lvardt", snames)
    chk("cv.states lvardt", s, tol=1e-8)
    cv.dstates(ds)
    chk("cv.dstates lvardt", ds, tol=1e-8)
    cv.error_weights(vec)
    chk("cv.error_weights lvardt", vec)
    cv.acor(vec)
    chk("cv.acor lvardt", vec, tol=5e-7)
    h.stoprun = 1
    cv.solve()
    cv.use_local_dt(0)
    expect_err("cv.statename(1e9, sref)")
    cv.active(0)
    expect_err("cv.statename(0, sref)")
    assert vec.size() > 0
    cv.error_weights(vec)
    assert vec.size() == 0
    vec.resize(20)
    cv.acor(vec)
    assert vec.size() == 0
    cv.active(1)

    del net, sref, snames, s, ds, std
    locals()


def state_magnitudes():
    h.load_file("stdrun.hoc")
    h.load_file("atoltool.hoc")
    net = Net(2)
    atool = h.AtolTool()
    h.tstop = 1

    def run():
        xtra = " lvardt" if cv.use_local_dt() else ""
        atool.anrun()
        r = [[i.name, i.max, i.acmax] for i in atool.states]
        chk("AtolTool" + xtra, r, tol=5e-7)
        h.init()
        atool.activate(1)
        h.run()
        cv.state_magnitudes(2)
        vec = h.Vector()
        cv.state_magnitudes(vec, 0)
        # bug for lvardt. slot for second cell is all 0
        chk("state_magnitudes states" + xtra, vec, tol=5e-8)
        cv.state_magnitudes(vec, 1)
        chk("state_magnitudes acor" + xtra, vec, tol=5e-7)
        cv.state_magnitudes(0)

    run()
    cv.use_local_dt(1)
    run()
    cv.use_local_dt(0)


def vec_record_discrete():
    net = Net(2)
    vec = h.Vector()
    trecord = h.Vector()
    tvec = h.Vector().indgen(0.1, 0.8, 0.1)
    vec.record(net.cells[0].soma(0.5)._ref_v, tvec, sec=net.cells[0].soma)
    trecord.record(h._ref_t, tvec, sec=net.cells[0].soma)
    cv.active(1)

    def run(tstop):
        pc.set_maxstep(10)
        h.finitialize(-65)
        h.frecord_init()
        pc.psolve(tstop)

    def ssrun(tstop):
        ss = h.SaveState()
        run(tstop / 2.0)
        ss.save()
        h.finitialize(-65)
        ss.restore()
        pc.psolve(tstop)

    run(1)
    chk("record discrete tvec", vec, tol=1e-9)
    tvec.indgen(1.1, 1.8, 0.1)
    ssrun(2)
    chk("record discrete savestate tvec", vec, tol=5e-7)
    cv.record_remove(vec)
    vec.record(net.cells[0].soma(0.5)._ref_v, 0.1, sec=net.cells[0].soma)
    trecord.record(h._ref_t, 0.1, sec=net.cells[0].soma)
    run(1)
    chk("record discrete dt", vec, tol=5e-9)
    ssrun(2)
    chk("record discrete savestate dt", vec, tol=5e-7)


def integrator_properties():
    s = h.Section()
    s.insert("pas")
    s.e_pas = 0
    s.g_pas = 0.001
    cv.active(1)
    vvec = h.Vector()
    tvec = h.Vector()
    vvec.record(s(0.5)._ref_v, sec=s)
    tvec.record(h._ref_t, sec=s)

    # rtol and atol
    def run1(key):
        h.finitialize(0.001)
        cv.solve(2)
        chk(key + " tvec", tvec, tol=5e-13)
        chk(key + " vvec", vvec, tol=1e-12)

    cv.rtol(1e-3)
    cv.atol(0)
    run1("rtol=%g atol=%g" % (cv.rtol(), cv.atol()))
    cv.rtol(0)
    cv.atol(1e-3)
    run1("rtol=%g atol=%g" % (cv.rtol(), cv.atol()))
    del s

    net = Net(2)
    vvec.record(net.cells[0].soma(0.5)._ref_v, sec=net.cells[0].soma)
    tvec.record(h._ref_t, sec=net.cells[0].soma)

    # stiff
    def run2(key):
        h.finitialize(-65)
        cv.solve(2)
        # Note the very large tolerance, this seems relatively unstable with the
        # NVIDIA compilers
        chk(key + " tvec size", tvec.size(), tol=0.25)

    cv.use_local_dt(1)
    cv.stiff(0)
    run2("stiff=0 lvardt")
    cv.stiff(2)
    cv.use_local_dt(0)
    for stiff in range(3):
        cv.stiff(stiff)
        run2("stiff=%d" % (cv.stiff(),))

    # maxorder, minstep, maxstep
    for lvardt in [1, 0]:
        cv.use_local_dt(lvardt)
        for maxorder in [2, 5]:
            cv.maxorder(maxorder)
            assert cv.maxorder() == maxorder
            h.finitialize(-65)
            while h.t < 2.0:
                cv.solve()
                order = cv.order(0) if lvardt else cv.order()
                assert order <= maxorder
        for minstep in [0.0001, 0.0]:
            cv.minstep(minstep)
            assert cv.minstep() == minstep
        for maxstep in [0.5, 1e9]:
            cv.maxstep(maxstep)
            assert cv.maxstep() == maxstep

    # jacobian
    for jac in [2, 1, 0]:  # end up as default
        cv.jacobian(jac)
        run2("jacobian " + str(cv.jacobian()))

    # selfqueue remove
    cv.active(0)
    cv.queue_mode(True, True)
    pc.set_maxstep(10)
    h.finitialize(-65)
    pc.psolve(2)
    cv.queue_mode(False, False)  # or later cv.store_events will abort(0)


def event_queue():
    # two cells with 1 and 3 connections to third cell
    def mknet():
        cells = {i: h.IntFire2() for i in range(3)}
        tvec = h.Vector()
        idvec = h.Vector()
        for i, cell in cells.items():
            cell.ib = 1.5 - 0.1 * i
            h.NetCon(cell, None).record(tvec, idvec, i)
        ncs = {}
        for i, numnc in enumerate([2, 3]):
            for j in range(numnc):
                nc = h.NetCon(cells[i], cells[2])
                ncs[(i, j)] = nc
                nc.weight[0] = -0.01
                nc.delay = 3
        return (cells, ncs, tvec, idvec)

    net = mknet()
    vecstore = h.Vector()
    cv.store_events(vecstore)
    h.finitialize()
    cv.solve(13)  # two NetCon from fast cell 0 and 1 presyn from slower cell 1
    cv.store_events()
    chk("store_events", vecstore, tol=1e-14)
    chk("event queue spikes", [net[2].to_python(), net[3].to_python()], tol=5e-16)
    old_stdout = sys.stdout
    sys.stdout = mystdout = io.StringIO()
    cv.print_event_queue()
    sys.stdout = old_stdout
    chk("print_event_queue", mystdout.getvalue(), tol=1e-14)
    tvec = h.Vector()
    cv.print_event_queue(tvec)
    chk("print_event_queue tvec", tvec, tol=1e-14)
    objs = h.List()
    flagvec = h.Vector()
    cv.event_queue_info(2, tvec, objs)
    chk(
        "event_queue_info(2...)",
        [tvec.to_python(), [o.hname() for o in objs]],
        tol=5e-16,
    )
    cv.event_queue_info(3, tvec, flagvec, objs)
    chk(
        "event_queue_info(3...)",
        [tvec.to_python(), flagvec.to_python(), [o.hname() for o in objs]],
        tol=5e-16,
    )

    # some savestate coverage
    def run(tstop):
        h.finitialize()
        cv.solve(tstop)

    run(25)

    def raster():
        print("spiketimes")
        net[2].printf()
        print("ids")
        net[3].printf()

    run(13)
    ss = h.SaveState()
    ss.save()
    sf = h.File("ss.bin")
    ss.fwrite(sf)
    del ss
    h.finitialize()
    ss = h.SaveState()
    ss.fread(sf)
    ss.restore()
    cv.solve(25)
    chk("SaveState restore at 13", [net[2].to_python(), net[3].to_python()], tol=5e-16)


def scatter_gather():
    net = Net(1)
    h.finitialize(-65)
    cv.solve(2)
    svec = h.Vector()
    cv.states(svec)
    yvec = h.Vector()
    cv.ygather(yvec)
    assert yvec.eq(svec)
    h.finitialize(-65)
    cv.yscatter(yvec)
    cv.re_init()
    cv.states(svec)
    assert yvec.eq(svec)
    cv.use_local_dt(1)
    expect_err("cv.yscatter(yvec)")
    expect_err("cv.ygather(yvec)")
    cv.use_local_dt(0)
    pc.nthread(2)
    expect_err("cv.yscatter(yvec)")
    expect_err("cv.ygather(yvec)")
    pc.nthread(1)
    yvec.resize(0)
    expect_err("cv.yscatter(yvec)")
    del yvec, svec, net
    locals()


def playrecord():
    net = Net(1)
    net.ic.dur = 1.0
    tvec = h.Vector([0, 1, 1])
    avec = h.Vector([0, 0.1, 0])
    avec.play(net.ic._ref_amp, tvec)
    cv.active(1)
    cv.debug_event(1)
    old_stdout = sys.stdout
    sys.stdout = mystdout = io.StringIO()
    h.finitialize(-65)
    cv.solve(1)
    cv.debug_event(0)
    sys.stdout = old_stdout
    chk("playrecord debug_event", debug_event_filter(mystdout.getvalue()))
    del net.ic
    h.finitialize(-65)


def interthread():
    net = Net(4)
    cv.active(0)
    h.dt = 0.025
    cv.debug_event(1)
    pc.nthread(2, 0)  # serial threads avoid race on print callback
    old_stdout = sys.stdout
    sys.stdout = mystdout = io.StringIO()
    h.finitialize(-65)
    while h.t < 10:
        h.fadvance()
    sys.stdout = old_stdout
    chk("interthread debug_event 2 serial threads", mystdout.getvalue())
    pc.nthread(1)
    cv.debug_event(0)


def test_netcvode_cover():
    nrn_use_daspk()
    node()
    netcon_access()
    cvode_meth()
    state_magnitudes()
    vec_record_discrete()
    integrator_properties()
    event_queue()
    scatter_gather()
    playrecord()
    interthread()


if __name__ == "__main__":
    test_netcvode_cover()
    chk.save()
    pass



"""
Serial batch implementation
for stimulus current i iterating over a range of values
run a simulation and report spike frequency
save i & corresponding f to a file
optionally plot fi curve
"""

from neuron import h, gui
import time

# Simulation parameters

PLOTRESULTS = True  # if True, generates a graph that shows f-i curve
h.tstop = 500  # ms, more than long enough for 15 spikes at ISI = 25 ms
AMP0 = 0.1  # nA -- minimum stimulus
D_AMP = 0.02  # nA -- stim increment between runs
NRUNS = 30

# model specification
from cell import Cell

cell = Cell()

# instrumentation

# experimental manipulations
stim = h.IClamp(cell.soma(0.5))
stim.delay = 1  # ms
stim.dur = 1e9
stim.amp = 0.1  # nA


def fstimamp(run_id):
    """takes run_id, returns corresponding stimulus amplitude"""
    return AMP0 + run_id * D_AMP


def set_params(run_id):
    """sets stimulus amplitude to appropriate value for this run"""
    stim.amp = fstimamp(run_id)


# data recording and analysis

# count only those spikes that get to distal end of dend
nc = h.NetCon(cell.dend(1)._ref_v, None, sec=cell.dend)
nc.threshold = -10  # mV
spvec = h.Vector()
nc.record(spvec)

NSETTLE = 5  # ignore the first NSETTLE ISI (allow freq to stablize)
NINVL = 10  # num ISI from which frequency will be calculated
NMIN = NSETTLE + NINVL  # ignore recordings with fewer than this num of ISIs


def get_frequency(spvec):
    nspikes = spvec.size()
    if nspikes > NMIN:
        t2 = spvec[-1]  # last spike
        t1 = spvec[-(1 + NINVL)]  # NINVL prior to last spike
        return NINVL * 1.0e3 / (t2 - t1)
    else:
        return 0


#
# Simulation control
#

# batch control
t_start = time.time()  # keep track of execution time past this point


def batchrun(n):
    stims = []
    freqs = []
    for run_id in range(n):
        set_params(run_id)
        h.run()
        stims.append(stim.amp)
        freqs.append(get_frequency(spvec))
        print("Finished %d of %d." % (run_id + 1, n))
    return stims, freqs


stims, freqs = batchrun(NRUNS)

#
# Reporting of results
#


def save_results(filename, stims, freqs):
    with open(filename, "w") as f:
        f.write("label:%s\n%d\n" % ("f", len(freqs)))
        for stim, freq in zip(stims, freqs):
            f.write("%g\t%g\n" % (stim, freq))


save_results("fi.dat", stims, freqs)

print("%g seconds" % (time.time() - t_start))  # report run time

if PLOTRESULTS:
    from plotfi import plotfi

    fig = plotfi(stims, freqs)



#!/usr/bin/env python

## Test of ParallelTransfer-based gap junctions. Assumes the presence
## of a conductance-based half-gap junction model ggap.mod

import sys, os, itertools, argparse, time
import numpy as np
from neuron import h


cells = []
gjlist = []
vrecs = []
stims = []


class MyCell:
    _ids = itertools.count(0)

    def __repr__(self):
        return "MyCell[%d]" % self.id

    def __init__(self):
        self.id = next(self._ids)
        # create the morphology and connect it
        self.soma = h.Section(name="soma", cell=self)
        self.dend = h.Section(name="dend", cell=self)
        self.dend.connect(self.soma(0.5))
        self.soma.insert("pas")
        self.dend.insert("pas")
        self.dend(0.5).pas.e = -65
        self.soma(0.5).pas.e = -65


## Creates half-gap junction mechanism
def mkgap(pc, sec, gid, secpos, sgid, dgid, w, gjlist):
    myrank = int(pc.id())

    seg = sec(secpos)
    gj = h.ggap(seg)
    gj.g = w

    pc.source_var(seg._ref_v, sgid, sec=sec)
    pc.target_var(gj, gj._ref_vgap, dgid)

    if myrank == 0:
        print(
            "mkgap: gid %i: sec=%s sgid=%i dgid=%i w=%f"
            % (gid, str(sec), sgid, dgid, w)
        )

    gjlist.append(gj)

    return gj


def mkcells(pc, ngids):
    nranks = int(pc.nhost())
    myrank = int(pc.id())

    assert nranks <= ngids

    for gid in range(ngids):

        if gid % nranks == myrank:

            cell = MyCell()
            nc = h.NetCon(cell.soma(0.5)._ref_v, None, sec=cell.soma)
            pc.set_gid2node(gid, myrank)
            pc.cell(gid, nc, 1)
            cells.append(cell)

            # Current injection into section
            stim = h.IClamp(cell.soma(0.5))
            if gid % 2 == 0:
                stim.delay = 10
            else:
                stim.delay = 20
            stim.dur = 20
            stim.amp = 10
            stims.append(stim)

            # Record membrane potential
            v = h.Vector()
            v.record(cell.dend(0.5)._ref_v)
            vrecs.append(v)

            if myrank == 0:
                print(
                    "Rank %i: created gid %i; stim delay = %.02f"
                    % (myrank, gid, stim.delay)
                )


## Creates gap junctional connections:
## The first halfgap is created on even gids and gid 0 and the second
## halfgap is created on odd gids.
def mkgjs(pc, ngids):
    nranks = int(pc.nhost())
    myrank = int(pc.id())

    ggid = 2e6  ## gap junction id range is intended to not overlap with gid range
    for gid in range(0, ngids, 2):

        # source gid: all even gids
        src = gid
        # destination gid: all odd gids
        dst = gid + 1

        # source gap junction gid
        sgid = ggid
        # destination gap junction gid
        dgid = ggid + 1

        # if the source gid exists on this rank, create the half gap from src to dst
        if pc.gid_exists(src) > 0:
            cell = pc.gid2cell(src)
            sec = cell.dend
            ggap = mkgap(pc, sec, gid, 0.5, sgid, dgid, 1.0, gjlist)

        # if the destination gid exists on this rank, create the half gap from dst to src
        if pc.gid_exists(dst) > 0:
            cell = pc.gid2cell(dst)
            sec = cell.dend
            ggap = mkgap(pc, sec, gid, 0.5, dgid, sgid, 1.0, gjlist)

        ggid += 2


def main():

    parser = argparse.ArgumentParser(description="Parallel transfer test.")
    parser.add_argument(
        "--sparse-partrans",
        dest="sparse_partrans",
        default=False,
        action="store_true",
        help="use sparse parallel transfer",
    )
    parser.add_argument(
        "--result-prefix",
        default=".",
        help="place output files in given directory (must exist before launch)",
    )
    parser.add_argument(
        "--ngids", default=2, type=int, help="number of gids to create (must be even)"
    )

    args, unknown = parser.parse_known_args()

    pc = h.ParallelContext()
    myrank = int(pc.id())

    mkcells(pc, args.ngids)
    mkgjs(pc, args.ngids)

    pc.setup_transfer()

    if args.sparse_partrans:
        if hasattr(h, "nrn_sparse_partrans"):
            h.nrn_sparse_partrans = 1

    rec_t = h.Vector()
    rec_t.record(h._ref_t)

    wt = time.time()

    h.dt = 0.25
    pc.set_maxstep(10)
    h.finitialize(-65)
    pc.psolve(500)

    total_wt = time.time() - wt

    gjtime = pc.vtransfer_time()

    print("rank %d: parallel transfer time: %.02f" % (myrank, gjtime))
    print("rank %d: total compute time: %.02f" % (myrank, total_wt))

    output = itertools.chain(
        [np.asarray(rec_t.to_python())],
        [np.asarray(vrec.to_python()) for vrec in vrecs],
    )
    np.savetxt(
        "%s/ParGJ_%04i.dat" % (args.result_prefix, myrank),
        np.column_stack(tuple(output)),
    )

    pc.runworker()
    pc.done()

    h.quit()


main()



# Basically want to test that net_move statement doesn't get
# mixed up with other instances.
import distutils.util
import os

from neuron import h

h.load_file("stdrun.hoc")

pc = h.ParallelContext()
h.dt = 1.0 / 4
h.steps_per_ms = 4


class Cell:
    def __init__(self, gid):
        self.soma = h.Section(name="soma", cell=self)
        if gid % 2 == 0:
            # CoreNEURON permutation not the identity if cell topology not homogeneous
            self.dend = h.Section(name="dend", cell=self)
            self.dend.connect(self.soma(0.5))
        self.gid = gid
        pc.set_gid2node(gid, pc.id())
        self.r = h.Random()
        self.r.Random123(gid, 0, 0)
        self.syn = h.DAsyn(self.soma(0.5))
        pc.cell(gid, h.NetCon(self.soma(0.5)._ref_v, None, sec=self.soma))
        # random start times for the internal events
        self.syn.tau0 = 10 * self.r.uniform(0.9, 1.1)
        self.syn.tau1 = 3 * self.r.uniform(0.9, 1.1)
        self.syn.dur = 6 * self.r.uniform(1, 10)

        self.ns = h.NetStim()
        self.ns.noise = 1
        self.ns.start = 0
        self.ns.number = 1e9
        self.ns.interval = 6
        self.ns.noiseFromRandom123(gid, 0, 1)
        self.nc = h.NetCon(self.ns, self.syn)
        self.nc.weight[0] = 0.1 * self.r.uniform(0.9, 1.1)
        self.nc.delay = 1
        self.msgvec = h.Vector()
        self.msgvec.record(self.syn._ref_msg, sec=self.soma)

    def result(self):
        return (self.syn.n_netsend, self.syn.n_netmove, self.msgvec.c())


def test_netmove():
    from neuron import coreneuron

    coreneuron.enable = False

    ncell = 10
    gids = range(pc.id(), ncell, pc.nhost())  # round robin

    cells = [Cell(gid) for gid in gids]

    tstop = 100

    def run(tstop, mode):
        pc.set_maxstep(10)
        h.finitialize(-65)
        if mode == 0:
            pc.psolve(tstop)
        elif mode == 1:
            while h.t < tstop:
                pc.psolve(h.t + 1.0)
        else:
            while h.t < tstop:
                h.continuerun(h.t + 0.5)
                pc.psolve(h.t + 0.5)

    run(tstop, 0)  # NEURON run

    stdlist = [cell.result() for cell in cells]

    print("CoreNEURON run")
    h.CVode().cache_efficient(1)
    coreneuron.enable = True
    coreneuron.verbose = 0
    coreneuron.gpu = bool(
        distutils.util.strtobool(os.environ.get("CORENRN_ENABLE_GPU", "false"))
    )

    def runassert(mode):
        run(tstop, mode)
        for i, cell in enumerate(cells):
            result = cell.result()
            std = stdlist[i]
            for j in range(2):
                if std[j] != result[j]:
                    print(
                        "cell {} std[{}]={} result[{}]={}".format(
                            i, j, std[j], j, result[j]
                        )
                    )
                    assert False
            if not std[2].eq(result[2]):
                print("Cell", i, "mode", mode)
                print(std[2], len(std[2]), result[2], len(result[2]))
                print("NEURON")
                print(list(std[2]))
                print("CoreNEURON")
                print(list(result[2]))
                assert False

    for mode in [0, 1, 2]:
        runassert(mode)

    coreneuron.enable = False
    # teardown
    pc.gid_clear()
    return stdlist


if __name__ == "__main__":
    from neuron import gui

    stdlist = test_netmove()
    g = h.Graph()
    print("n_netsend  n_netmove")
    for result in stdlist:
        print(result[0], result[1])
        result[2].line(g)
    g.exec_menu("View = plot")
    h.quit()



from . import ctng
from . import scalarField
import numpy
from .surfaces import chunkify
from . import surfaces
from numpy.linalg import norm

_max_chunks = 10000000

# TODO: should we expand this to also compute the area between neighboring nodes?


def voxelize2(
    source,
    dx=0.25,
    xlo=None,
    xhi=None,
    ylo=None,
    yhi=None,
    zlo=None,
    zhi=None,
    n_soma_step=100,
    relevant_pts=None,
):
    """
    Generates a cartesian mesh of the volume of a neuron, together with discretized information on
    surface areas and volumes. This is more accurate than voxelize, which only checks the center
    point of a voxel, but it is slower.

    Parameters
    ----------
    source : :func:`list`, ``nrn.SectionList``, or ``nrn.Import3D``
        The geometry to mesh.
    dx : double, optional
        Mesh step size.
    xlo : double, optional
        Minimum x value. If omitted or None, uses minimum x value in the geometry.
    xhi : double, optional
        Maximum x value. If omitted or None, uses maximum x value in the geometry.
    ylo : double, optional
        Minimum y value. If omitted or None, uses minimum y value in the geometry.
    yhi : double, optional
        Maximum y value. If omitted or None, uses maximum y value in the geometry.
    zlo : double, optional
        Minimum z value. If omitted or None, uses minimum z value in the geometry.
    zhi : double, optional
        Maximum z value. If omitted or None, uses maximum z value in the geometry.
    n_soma_step : integer, optional
        Number of pieces to slice a soma outline into.

    relevant_pts : list, optional
        A list for each section in sources with a list of the 3D points in that section that should be used.

    Returns
    -------
    mesh : :class:`ScalarField`
        The mesh. Values are scalars, but may be used as True inside the
        geometry and False outside.
    surface_areas : :class:`ScalarField`
        The total surface area passing through the node.
    volumes : :class:`ScalarField`
        The volume of the neuron contained in the given node.

    Examples
    --------

    Basic usage:

    >>> mesh, surface_areas, volumes = geometry3d.voxelize2(h.allsec())

    Full example, using :mod:`pyplot`:

    >>> s1, s2, s3 = [h.Section() for i in xrange(3)]
    >>> for sec in [s2, s3]: ignore_return = sec.connect(s1)
    ...
    >>> for sec in h.allsec():
    ...     sec.diam = 1
    ...     sec.L = 5
    ...
    >>> mesh = geometry3d.voxelize2(h.allsec(), dx=.1)[0]
    >>> for i in xrange(10):
    ...     ignore_return = pyplot.subplot(2, 5, i + 1)
    ...     ignore_return = pyplot.imshow(mesh.values[:, :, i])
    ...     ignore_return = pyplot.xticks([])
    ...     ignore_return = pyplot.yticks([])
    ...
    >>> pyplot.show()

    .. plot::

        from neuron import h
        from matplotlib import pyplot
        from . import geometry3d

        s1, s2, s3 = [h.Section() for i in xrange(3)]
        for sec in [s2, s3]: ignore_return = sec.connect(s1)

        for sec in h.allsec():
            sec.diam = 1
            sec.L = 5

        mesh = geometry3d.voxelize2(h.allsec(), dx=.1)[0]
        for i in xrange(10):
            ignore_return = pyplot.subplot(2, 5, i + 1)
            ignore_return = pyplot.imshow(mesh.values[:, :, i])
            ignore_return = pyplot.xticks([])
            ignore_return = pyplot.yticks([])

        pyplot.show()



    .. note::
        The use of Import3D objects is recommended over lists of sections
        because the former preserves the soma outline information while
        the later does not. Up to one soma outline is currently supported.
    """

    objects = ctng.constructive_neuronal_geometry(
        source, n_soma_step, dx, relevant_pts=relevant_pts
    )

    if xlo is None:
        xlo = min(obj.xlo for obj in objects) - 3 * dx
    if ylo is None:
        ylo = min(obj.ylo for obj in objects) - 3 * dx
    if zlo is None:
        zlo = min(obj.zlo for obj in objects) - 3 * dx
    if xhi is None:
        xhi = max(obj.xhi for obj in objects) + 3 * dx
    if yhi is None:
        yhi = max(obj.yhi for obj in objects) + 3 * dx
    if zhi is None:
        zhi = max(obj.zhi for obj in objects) + 3 * dx

    mesh = scalarField.ScalarField(xlo, xhi, ylo, yhi, zlo, zhi, dx, dtype="B")
    grid = mesh.values

    xs, ys, zs = mesh.xs, mesh.ys, mesh.zs

    # use chunks no smaller than 10 voxels across, but aim for max_chunks chunks
    chunk_size = max(
        10, int((len(mesh.xs) * len(mesh.ys) * len(mesh.zs) / _max_chunks) ** (1 / 3.0))
    )

    chunk_objs, nx, ny, nz = chunkify(
        objects, mesh.xs, mesh.ys, mesh.zs, chunk_size, dx
    )

    # this is the crude volumetric approach shared with voxelize
    for i, x in enumerate(mesh.xs):
        chunk_objsa = chunk_objs[i // chunk_size]
        for j, y in enumerate(mesh.ys):
            chunk_objsb = chunk_objsa[j // chunk_size]
            for k, z in enumerate(mesh.zs):
                grid[i, j, k] = is_inside(x, y, z, chunk_objsb[k // chunk_size])

    surface_areas = scalarField.ScalarField(xlo, xhi, ylo, yhi, zlo, zhi, dx)
    sa_grid = surface_areas.values

    # triangulate the surface
    triangles = surfaces._triangulate_surface_given_chunks(
        objects, xs, ys, zs, False, chunk_size, chunk_objs, nx, ny, nz, True, sa_grid
    )

    # for each triangle, compute the area. Add it to the appropriate spots in sa_grid
    # TODO: move this to C? or at least cythonize it?
    for tdata in triangles.reshape(len(triangles) // 9, 9):
        v0, v1, v2 = tdata[0:3], tdata[3:6], tdata[6:9]
        centerx, centery, centerz = (v0 + v1 + v2) / 3
        i, j, k = (centerx - xlo) / dx, (centery - ylo) / dx, (centerz - zlo) / dx
        sa_grid[i, j, k] += 0.5 * norm(numpy.cross(v1 - v0, v2 - v0))

    # now ensure that any grid containing surface is included in the voxelization
    # TODO: change this when supporting multiple non-overlapping regions in one 3d Domain
    grid[sa_grid.nonzero()] = 1

    volumes = scalarField.ScalarField(xlo, xhi, ylo, yhi, zlo, zhi, dx)
    volume_values = volumes.values

    # TODO: compute the correct partial volumes for nodes on the boundary

    # start out with the full volumes for every place that contains volume
    volume_values[grid.nonzero()] = dx**3

    # correct the boundary node volumes
    # surface_nodes = sa_grid.nonzero()
    # for i, j, k in zip(*surface_nodes):
    #    volume_values[i, j, k] = surfaces.volume_inside_cell(i, j, k, chunk_objs[i // chunk_size][j // chunk_size][k // chunk_size], xs, ys, zs)

    return mesh, surface_areas, volumes, triangles.reshape(len(triangles) // 9, 9)


# inside the neuron if inside of any of its parts
def is_inside(x, y, z, active_objs):
    return 1 if any(obj(x, y, z) <= 0 for obj in active_objs) else 0



import numpy
from neuron import h

_h_vector = h.Vector
_numpy_zeros = numpy.zeros
_h_ptrvector = h.PtrVector


def _donothing():
    pass


class RangeVar:
    def __init__(self, name):
        self._name = name
        self._ptr_vector = None

    def _init_ptr_vectors(self, nodes):
        # TODO: make sure this gets changed everytime there is a structurechange event
        #       or everytime the nodes change
        ptrs = []
        locs = []
        ptrs_append = ptrs.append
        locs_append = locs.append
        name = self._name
        for node in nodes:
            seg = node.segment
            ptrs_append(getattr(seg, "_ref_%s" % name))
            # TODO: is this the right index? or do I need to change things to
            #       account for the zero-volume nodes?
            locs_append(node._index)
        self._locs = numpy.array(locs)
        pv = _h_ptrvector(len(ptrs))
        pv.ptr_update_callback(_donothing)
        pv_pset = pv.pset
        for i, ptr in enumerate(ptrs):
            pv_pset(i, ptr)
        self._pv = pv

    def _rangevar_vec(self):
        pv = self._pv
        result = _numpy_zeros(pv.size())
        vec = _h_vector(pv.size())
        pv.gather(vec)
        vec.to_python(result)
        return result



import pytest


def test_region_sections(neuron_instance):
    """A test that regions and species do not keep sections alive"""

    h, rxd, data, save_path = neuron_instance
    soma = h.Section(name="soma")
    soma.L = 5
    soma.diam = 5
    cyt = rxd.Region(soma.wholetree(), name="cyt", dx=1)
    ca = rxd.Species(cyt, name="ca", charge=2)
    h.finitialize(-65)
    soma = None

    for sec in h.allsec():
        assert False


def test_section_removal(neuron_instance):
    """A test that deleted sections are removed from rxd"""

    h, rxd, data, save_path = neuron_instance
    soma = h.Section(name="soma")
    soma.L = 5
    soma.diam = 5
    rxd.set_solve_type(soma, dimension=3)
    cyt = rxd.Region(soma.wholetree(), name="cyt", dx=1)
    ca = rxd.Species(cyt, name="ca", charge=2)
    h.finitialize(-65)

    soma = h.Section(name="soma")
    h.finitialize(-65)

    assert not ca.nodes
